#include "../h/pmcIO.h"
#include "pmc_pid.h"
#include "../h/cdefs.h"
#include "mc8prim.h"
#include "../h/moper_com.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include <sys/types.h>
#include <fcntl.h>
#include <strings.h>

#ifdef X86sol_TARGET
#include "pmcioctl.h"
#endif

PID rciPmcPid;
PIDHEADER rciPmcPidHeader;

#define SERVO_MODE

typedef struct {
	int nchan;
	int *fd;
	int *chan;
	int *mode;
	mc8_struct *pmc8;
} PmcInfo;

union pack3 {
	short s[4];
	int i[2];
  };

static void writeInt (PmcInfo *pmci, int addr, int val)
{
	union pack3 xfer;
	
	xfer.i[0] = val;
	mc8WriteMem (pmci->pmc8, addr, xfer.s, /*numWords=*/1);
}

static int readInt (PmcInfo *pmci, int addr)
{
	union pack3 xfer;

	mc8ReadMem (pmci->pmc8, addr, xfer.s, /*numWords=*/1);
	return xfer.i[0];
}

void static doclose (PmcInfo *pmci)
{
	int k;
	
	if (pmci->fd)
	 { for (k=0; k<pmci->nchan; k++)
	    { if (pmci->fd[k] != -1)
	       { close (pmci->fd[k]);
	       }
	    }
	   free (pmci->fd);
	 }
	if (pmci->chan)
	 { free (pmci->chan);
	 }
	if (pmci->mode)
	 { free (pmci->mode);
	 }
	free (pmci);
}

int pmcio_attach (IO_DEVICE *dev, int nchan)
{
	char exopenDevName[256];
	char *fxnName = "pmcio_attach()";
	char unitChar;
	char *cp;
	PmcInfo *pmci;
	int k;

	pmcio_detach ( dev ); /* XXX XXX XXX!!! */

	pmci = (PmcInfo*)malloc(sizeof(PmcInfo));
	pmci->fd = (int*)malloc(sizeof(int)*nchan);
	pmci->chan = (int*)malloc(sizeof(int)*nchan);
	pmci->mode = (int*)malloc(sizeof(int)*nchan);
	for (k=0; k<nchan; k++)
	 { pmci->fd[k] = -1;
	 }
	pmci->nchan = nchan;

	/* open the exclusive-open devices for each of the three
	   channels */
	printf( "Device Name: %s   Unit: %d\n", dev->name, dev->unit );
	if ( strncmp( "pmc", dev->name, 3 ) == 0 ) {
	  sprintf (exopenDevName, "/dev/%sChan?_%d", "pmc", dev->unit);
	} else {
	  sprintf (exopenDevName, "/dev/%sChan?_%d", dev->name, dev->unit);
	}
	cp = strchr (exopenDevName, '?');

	for (k=0; k<nchan; k++)
	 { pmci->chan[k] = dev->args[k];
	   *cp = pmci->chan[k] + '0';
	   if ((pmci->fd[k] = open (exopenDevName, O_RDWR)) == -1)
	    { errorMsg (EConnect, "%s: can't open channel %d\n", fxnName, k);
	      sysErrorMsg (EConnect, fxnName, "open");
	      doclose (pmci);
	      return -1;
	    }
	 }
#ifdef X86sol_TARGET
	/* Do the ioctl here to enable user mode access to I/O space.
	   We can use any of the open file descriptors. */
	if (ioctl (pmci->fd[0], PMCIOC_ENABLE_IOPORTS, 1) == -1 ||
	    ioctl (pmci->fd[0], PMCIOC_ENABLE_IOPORTS, 0) == -1)
	 { sysErrorMsg (EConnect, fxnName, "ioctl PMCIOC_ENABLE_IOPORTS");
	   doclose (pmci);
	   return -1;
	 }
#endif
	if ((pmci->pmc8 = mc8Open(dev->address, 0, 0)) == NULL)
	 { errorMsg (EConnect, "%s: no response from PMC\n", fxnName);
	   doclose (pmci);
	   return -1;
	 }
	/* check magic and version number */
	k = readInt (pmci, PID_MAGIC_ADDR);
	if (k != PID_MAGIC)
	 { errorMsg (EConnect, "%s: bad magic number: 0x%x vs. 0x%x\n", 
				fxnName, k, PID_MAGIC);
	   doclose (pmci);
	   return -1;
	 }
	k = readInt (pmci, PID_VERSION_ADDR);
	if (k != PID_VERSION)
	 { errorMsg (EConnect, "%s: wrong version: %d vs. %d\n",
			       fxnName, k, PID_VERSION);
	   doclose (pmci);
	   return -1;
	 }
	for (k=0; k<nchan; k++)
	 { pmci->mode[k] = readInt (pmci, PID_MODE_ADDR(pmci->chan[k]));
	 }

	dev->location = (char*)pmci;	
	return 1;
}

void pmcio_detach (IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;

	if( pmci )
	{
	   mc8Close (pmci->pmc8);
	   doclose (pmci);
	   dev->location = 0;
	}
}

void pmcio_set_rate (IO_DEVICE *dev, int interval)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	int ctrlRate;
	
	ctrlRate = readInt (pmci, PID_CTRL_RATE_ADDR);
	writeInt (pmci, PID_INTERPOLATE_ADDR, 
			      (int)(interval*ctrlRate/1000.0+0.99));
}

void pmcio_set_power (IO_DEVICE *dev, int power)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
}

void pmcio_set_stop (IO_DEVICE *dev, int j)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
}

void pmcio_set_encvalue (IO_DEVICE *dev, int val, int j)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;

	writeInt (pmci, PID_POSN_ADDR(pmci->chan[j]), val);
}

void pmcio_set_enc (IO_DEVICE *dev, int val, int j)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;

	if (pmci->mode[j] != PID_SERVO_ENABLE)
	 { writeInt (pmci, PID_MODE_ADDR(pmci->chan[j]), PID_SERVO_ENABLE);
	   pmci->mode[j] = PID_SERVO_ENABLE;
	 }
	writeInt (pmci, PID_GOAL_ADDR(pmci->chan[j]), val);
}

void pmcio_set_cur (IO_DEVICE *dev, int cur, int j)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;

	if (pmci->mode[j] != 0)
	 { writeInt (pmci, PID_MODE_ADDR(pmci->chan[j]), 0);
	   pmci->mode[j] = 0;
	 }
}

void pmcio_set_integration (IO_DEVICE *dev, int mask)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
}

void pmcio_set_calibMode (IO_DEVICE *dev, int mask)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
}

int pmcio_get_state(IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	
	return IS_CALIBRATED | POWER_ON;	
}

int pmcio_encset0[12];
int pmcio_encset1[12];
int pmcio_chan[12];
int pmcio_numtrys;

#if 0
void pmcio_get_enc (IO_DEVICE *dev, int *enc)
{
#define MAX_TRYS	4
	PmcInfo *pmci = (PmcInfo*)dev->location;
	int buf[MAX_TRYS];
	int i, j, k, chan;
	static int done = 0;

	for (j=0; j<pmci->nchan; j++)
	 { k = 0;
	   chan = pmci->chan[j];
	   buf[0] = readInt (pmci, PID_POSN_ADDR(chan));
	   do
	    { buf[++k] = readInt (pmci, PID_POSN_ADDR(chan));
	    }
	   while (buf[k] != buf[k-1] && k < MAX_TRYS-1);
	   if (buf[k] != buf[k-1])
	    { printf ("Error: pmcio_get_enc(), device %s\n", dev->name);
	      printf ("inconsistent data, j=%d, after %d tries\n",j,MAX_TRYS);
	      for (i=0; i<=k; i++)
	       { printf ("%d %d\n", i, buf[i]);
	       }
	      exit (1);
	    }
	   if (k > 1)
	    { printf ("pmcio_get_enc: %d tries %8d %8d\n", k+1,
		       buf[0], buf[1]);
	    }
	   enc[j] = buf[k];
	   pmcio_encset0[j] = buf[0];
	   pmcio_encset1[j] = buf[1];
	   pmcio_chan[j] = chan;
	   pmcio_numtrys = k+1;
	 }
}
#else
void pmcio_get_enc (IO_DEVICE *dev, int *enc)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	int j;
	static int done = 0;

	for (j=0; j<pmci->nchan; j++)
	 { enc[j] = readInt (pmci, PID_POSN_ADDR(pmci->chan[j]));
	 }
}
#endif

void pmcio_get_mode (IO_DEVICE *dev, int *mode)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	int j;

	for (j=0; j<pmci->nchan; j++)
	 { mode[j] = readInt (pmci, PID_MODE_ADDR(pmci->chan[j]));
	 }
}

void pmcio_get_cur (IO_DEVICE *dev, int *rawTor)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	int k;

	for (k=0; k<pmci->nchan; k++)
	 { 
	   rawTor[k] = 0;
	 }
}

int pmcio_get_integration (IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	return 0;
}

int pmcio_get_calibMode (IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	return 0;
}

int pmcio_get_mark (IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	return 0;
}

int pmcio_get_options (IO_DEVICE *dev)
{
	PmcInfo *pmci = (PmcInfo*)dev->location;
	return 0;
}
