/*=-=-=-=-=-=-=-=-=-=-  RCCL/RCI Notice of Copyright  -=-=-=-=-=-=-=-=-=-=-=*
 *									    *
 *	Copyright (C) 1987 by all authors described below. Permission to    *
 *	use, distribute and copy is granted, but only for non-commercial    *
 *	purposes and providing the authors are acknowledged and this 	    *
 *	copyright notice is displayed prominently.			    *
 *									    *
 *	The authors provide this software in hope that it will be useful,   *
 *	but do not assume responsibility for its use.			    *
 *									    *
 *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *
 *
 * NAME:	down -- download programs to the PUMA controller
 *
 * USAGE:	down file [-a <address>] [-d <device>] [-e <entry>] [-t] [-s]
 *
 * DESCRIPTION:	This program is used to download PDP-11 programs from a UNIX
 *		host to a UNIMATION PUMA controller. The loaded programs
 *		execute in place of the VAL code.
 *
 *		The program works by openning a serial port connected to the
 *		LSI-11's console line. This port may be specified with the -d
 *		option; the default is "/dev/puma0", which may be set as a
 *		symbolic link to any tty port which is convenient. The port
 *		must have no 'getty' process on it. A baud rate of 9600 is
 *		assumed.
 *
 *		The program first tests the serial line to see if the LSI-11
 *		is in ODT. If it is not, it prompts the user to power up or
 *		reset the LSI-11 (which is assumablly running Unimation VAL
 *		software) and then sends the appropriate messages to VAL to
 *		get it to exit and enter ODT.
 *
 *		ODT is used to enter a very small bootstrap which in turn
 *		loads the main body of the code over the console line using
 *		full 8-bit wide transmission.
 *
 *		The command line contains arguments specifying the file to be
 *		loaded, the address to load to, and the starting entry point.
 *
 *		After the down loading is completed, the program continues to
 *		communicate with the LSI-11 over the serial line, unless the
 *		-t option has been specified. Communication is broken and the
 *		program exits when the break character '~' is typed.
 *
 * ARGUMENTS:	file	-- a file containg a runable LSI-11 image
 *
 *		-a -- 	specifiy load address (in octal)
 *
 *		-e --	specify entry point for program (in octal)
 *
 *		-d --	specify serial line to do the down loading over
 *
 *		-t --	terminate the program when the down loading is
 *			finished.
 *
 *		-s -- 	run the program without printing any information or
 *			messages.
 *
 *		-v --	Verbose mode. Print out all characters sent back back
 *			and forth between the VAX and VAL or ODT. This is used
 *			for debugging purposes.
 *
 *		-T --	WARNING: THIS FEATURE MAY BE A LITTLE BUGGY
 *			Tells down to turn the specified terminal line off in
 *			/etc/ttys before starting (and turn it back on again 
 *			when exiting).  This option will be ignored if 
 *			termlink is not installed set-uid or if the specified
 *			line is not listed in /etc/ttys.  If someone is
 *			logged in on the specified line, down will merely
 *			print a message and exit.
 *
 *		-lsi11_02
 *		-lsi11_23
 *		-lsi11_73
 *			Tell the program the type of LSI11 CPU that it is
 *			downloading to. The default is an lsi11_73.
 * 
 * FILES:	lsiboot --		Contains (in octal)  a primary
 *					bootstrap which down initially loads
 *					into the controller using ODT.
 *
 * CAVEATS:	The program is reasonably reliable, although the internals at
 *		this point resemble what Dr. Zucker calls a ``hospital ward''.
 *
 * AUTHOR/DATE:	John Lloyd, Fall 1984.
 *
 * UPDATES:	John Lloyd, Fall 1986: added feature whereby program checks to
 *		see if LSI-11 is already in ODT.
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <stdio.h>
#include <signal.h>
#include <sys/file.h>
#include <errno.h>
#include <utmp.h>
#include <ctype.h>

#ifdef HAS_SYSSELECT_H
#include <sys/select.h>
#endif

#include "../h/termline.h"

#ifdef HAS_SGTTY
#define TURN_OFF_SUPPORTED		1
#else
#define TURN_OFF_SUPPORTED		0
#endif

#define ARRAY_SIZE(a)	(sizeof(a)/sizeof(a[0]))   /* # of array elements */

#define READ 0			/* Codes for the "open" command		*/
#define READ_WRITE 2		
#define CHAR_SIZE 1		/* Size of 1 character			*/
#define BOOTSTART 077500	/* Address in LSI where bootstrap is put*/
#define TRUE 1
#define INTERRUPT_CHAR '~'	/* '~' typed by user to stop program	*/
#define HEADER_SIZE 16		/* size of header info in download file */
#define	BUFFER_SIZE 1024	/* size of data buffer for downloading  */
#define DEFAULT_START 010000	/* default load address			*/
#define MAXTRYS 10		/* Max number of times to retransmit	*/
#define GETCHAR_TIMEOUT	5	/* Number of seconds before timeout	*/

int tty_fd, line_fd;		/* channels for getting characters from */
				/* the terminal and the lsi		*/
int pid;			/* process id				*/
int silent = 0;			/* supresses printing of information	*/
int verbose = 0;		/* prints out all data going back and forth */
char boot_file[256] 		/* File containing LSI bootstrap	*/
    = "/local/lib/lsiboot";
char pdp_device [20]		/* Serial line connected to LSI		*/
    = "/dev/puma0";

int bigEndianHost;		/* Is this machine little of big endian? */

#define BUF_SIZE 256		/* size of the character buffer */
char buffer [BUF_SIZE];		/* character buffer for talking to LSI */



TermProps old_ttystate;
TermProps old_linestate;

escape (message)
char *message;
 {
   /*
    * Called to exit the program. Cleans up the user's terminal 
    * characteristics and returns control to the shell. Prints an optional
    * message to stderr.
    */

   termlineSet (tty_fd, &old_ttystate);
   termlineSet (line_fd, &old_linestate);

   if (message != NULL && !silent)
      fprintf (stderr, "%s", message);

   kill (pid, SIGINT);
   kill (getpid(),SIGINT);
 }


SIGHAN_RETURN timeout_handler(sig)
int sig;
{
 fprintf(stderr,"Timeout on LSI line\n");
 exit(1);
}

argcmp (arg, string)
char *arg, *string;
{
 /*
  * Matches string against arg. If arg is shorter than string but doesn't
  * contradict it, then the match is allowed; if the arg is the same as the
  * string, then of course the match is allowed, otherwise the match is not
  * allowed 
  */

 int succeed = 1;

 while ((*string != 0) && (*arg != 0) && succeed)
  { 
    if (*string != *arg)
       succeed = 0;
    else
     { 
       string++;
       arg++;
     }
  }
 if ((*string == 0) && (*arg != 0))
    succeed = 0;
 return (succeed);
}


strlist_match (arg, strlist, list_size)
char *arg;
char *strlist [];
int list_size;
{
 /* strlist_match checks to see if arg compares with any of the strings in
  * strlist. If it checks with none, -1 is returned. If it checks with
  * more than one, -2 is returned. If it checks with one, then the index
  * of the string that it checks with is returned. 
  */

 int index=0, match_count=0, last_match=0;
 
 while (index < list_size)
  { 
    if (argcmp (arg, strlist[index]))
     { 
       match_count++;
       last_match = index;
     }
    index++;
  }
 if (match_count == 1)
    return (last_match);
 else if (match_count > 1)
    return (-2);
 else /* match_count == 0 */
    return (-1);
}


get_lsi_char ()
 {
   /* Returns one character read from the lsi */

   char lsi_char;
   read (line_fd, &lsi_char, CHAR_SIZE);
   if (verbose)
    { putchar (lsi_char);
      fflush (stdout);
    }
   return (lsi_char);
 }


int wait_lsi_any(match)
char *match;
/* waits for one character of a string to be received from the lsi */
{
 int l = strlen(match);
 int i;
 char c;

alarm(GETCHAR_TIMEOUT);
 while (1)
  { c = get_lsi_char();
    for (i=0;i<l;i++)
     { if (c == match[i])
	{ alarm(0);
	  return(i);
	}
     }
  }
}

#include "../h/varargInc.h"

wait_lsi VA_ARGDEF ((char *arg1, ...))
 {
   /* waits for specific character strings to be sent by the lsi */

#if !USE_STDARG
   char *arg1;
#endif

#define IBUF_SIZE 1024
   static char iBuf[IBUF_SIZE];
   va_list ap;

   int bp = 0;
   int argnum, length;
   int match = 0;
   char *argv;

   alarm(GETCHAR_TIMEOUT);

   while (match == 0 && bp < IBUF_SIZE)
    { iBuf[bp++] = get_lsi_char();
      argnum = 1;
#if !USE_STDARG
      va_start(ap);
      arg1 = va_arg(ap, char*);
#else
      va_start(ap, arg1);
#endif
      for (argv = arg1; argv; argv = va_arg(ap, char*))
       { length = strlen(argv);
	 if (length <= bp)
	  { if (strncmp (&iBuf[bp-length], argv, length) == 0)
	     { match = argnum;
	     }
	  }
	 argnum++;
       }
      va_end(ap);
    }

   alarm (0);
   return (match);
 }


send_lsi (string)
 char * string;
 {

   /* sends down a string to the lsi, waiting for the echo each time */

   int index, string_length;
   char lsi_char;
   
   alarm(GETCHAR_TIMEOUT);
   string_length = strlen (string);
   for (index = 0; index < string_length; index++)
    { 
      lsi_char = string [index];
      write (line_fd, &lsi_char, CHAR_SIZE);
      read (line_fd, &lsi_char, CHAR_SIZE);
      if (verbose)
       { putchar (lsi_char);
	 fflush (stdout);
       }
    }
   alarm(0);
 }

put_lsi (string)
char *string;
{
   /* sends down a string to the lsi, but does not wait for echo */

   int index, string_length;
   char lsi_char;
   string_length = strlen (string);
   for (index = 0; index < string_length; index++)
    { 
      lsi_char = string [index];
      write (line_fd, &lsi_char, CHAR_SIZE);
    }
 }


read_lsi(buffer, bufSize, time)
char *buffer;
int bufSize;
float time;
{
	int n, count;

	count = 0;
	while (testinput (line_fd, time))
	 { n = read (line_fd, &buffer[count], bufSize - count);
	   count += n;
	 }
	buffer[count] = '\0';
	return (count);
}

check_lsi_sequence (buffer, numChars, string)
char *buffer, *string;
int numChars;
{
	int l = strlen(string);

	if (numChars < l)
	 { return (0);
	 }
	if (strncmp (&buffer[numChars-l], string, l) != 0)
	 { return (0);
	 }
	return (1);
}

load_value (address, value)
char *address, *value;
{
 	send_lsi (address);
	send_lsi ("/");
	wait_lsi (" ", 0);
	send_lsi (value);
	send_lsi ("\r");
	wait_lsi ("@", 0);
}



int old_active = -1;		    /* old activity state of transfer line */
char **argvec;

testForBigEndian()
{
	union {
	  unsigned short s;
	  unsigned char c[2];
	  } swap;

	swap.s = 1;
	return (swap.c[1] == 1);
}

unsigned short swapShort (s)
unsigned short s;
{
	union {
	  unsigned short s;
	  unsigned char c[2];
	  } swap;
	unsigned char tmp;

	swap.s = s;
	tmp = swap.c[0];
	swap.c[0] = swap.c[1];
	swap.c[1] = tmp;
	return (swap.s);
}

writeShort (fd, s)
int fd;
unsigned short s;
{
	if (bigEndianHost)
	 { s = swapShort(s);
	 }
	write (fd, &s, sizeof(short));
}

#define LSI11_02	1
#define LSI11_23	2
#define LSI11_73	3

int cpuType = LSI11_73;

#if !TURN_OFF_SUPPORTED
printNoTurnOff()
{
	fprintf(stderr, "\
The -T option is not supported on this system. Normally, this option turns\n\
off logins on the port that you wish to connect to, for the duration of the\n\
termlink session. This may not be necessary because logins are often not\n\
enabled on spare serial ports. However, if they are enabled on the port you\n\
wish to use, they can be turned off permanently by your system admistrator.\n\
");
}
#endif

main (argc, argv)
 int argc;
 char ** argv;
 {
  TermProps new_linestate;
  TermProps new_ttystate;
  char lsi_char;		/* character returned from lsi */
  char octal_num [6];		/* string to place an octal number */
  int to_fd;			/* file descriptor for transmitting */
  int from_fd;			/* file descriptor for receiving */
  int load_fd;			/* descriptor for file to be downloaded */
  char *filename = 0;		/* file to be downloaded */
  FILE *bootstrap;
  int debug = 0;
  int i;

  struct file
   {
     short	a_magic;
     short	a_text;
     short	a_data;
     short	a_bss;
     short	a_syms;
     short	a_entry;
     short	a_trsize;
     short	a_drsize;
   } loadhead;

  struct {			    /* Packet downloaded to LSI:	*/
     unsigned short nbytes;	    /*   number of bytes coming down 	*/
     unsigned char buffer[BUFFER_SIZE]; /*   bytes to be downloaded 	*/
   } loadbuf;

  int num_bytes;		    /* number of bytes to be downloaded */
  int start = -1;		    /* starting address in lsi to load */
  int entry = -1;		    /* entry point in the lsi */
  int num_packets;	    	    /* number of packets to send to lsi */
  int last_packet_size;  	    /* size of the last packet (in bytes) */
  unsigned short checksum;	    /* checksum information		*/
  int packet_count;		    /* current packet being sent	*/
  int num_trys;			    /* number of times sending packet   */
  int on = 0;			    /* true if the arm is on already */
  int terminate = 0;		    /* flag to terminate downloader */
  int qual;			    /* command qualifier number */
  int num_chars;		    /* number of characters read into buffer*/
  int turn_off = 0;		    /* turn off login process */ 

  static char *qualifiers[] = { "-entry",
  				"-address",
				"-on",
				"-terminate",
				"-silent",
				"-dev",
				"-verbose",
				"-T",
				"-lsi11_02",
				"-lsi11_23",
				"-lsi11_73"};
    
  short header [HEADER_SIZE/2];	/* recieves header for download file */

  /* test to see what sort of machine we have */

  bigEndianHost = testForBigEndian();

  /* parse the input line */

  argvec = argv;
  while (argc-- > 1)
    { 
      ++argv;
      if (**argv == '-') 	/* then arg is a qualifier */
       { 
	 if ((qual = strlist_match (*argv, qualifiers, 
					   ARRAY_SIZE(qualifiers))) >= 0)
	  { 
	    switch (qual)
	     { 
	       case 0:
		  argv++;
		  argc--;
		  if (sscanf (*argv, "%o", &entry) <= 0)
		     {
		       fprintf (stderr, "ILLEGAL QUALIFIER VALUE for entry \n");
		       exit (1);
		     }
		  break;
	       case 1:
		  argv++;
		  argc--;
		  if (sscanf (*argv, "%o", &start) <= 0)
		     {
		       fprintf (stderr, "ILLEGAL QUALIFIER VALUE for start \n");
		       exit (1);
		     }
		  break;
	       case 2:
		  on = 1;
		  break;
	       case 3:
		  terminate = 1;
		  break;
	       case 4:
		  silent = 1;
		  break;
	       case 5:
	          argv++;
	          argc--;
	          sscanf (*argv, "%19s", pdp_device);
	          break;
	       case 6:
		  verbose = 1;
		  break;
	       case 7:
#if TURN_OFF_SUPPORTED		  
		  turn_off = 1;
#else
		  printNoTurnOff();
		  exit (0);  
#endif
		  break;
	       case 8:
		  cpuType = LSI11_02;
		  break;
	       case 9:
		  cpuType = LSI11_23;
		  break;
	       case 10:
		  cpuType = LSI11_73;
		  break;
	     }
	  }
	 else if (qual == -1)
	  { 
	    fprintf (stderr,"ILLEGAL QUALIFIER %s\n", *argv);
	    exit (1);
	  }
	 else if (qual == -2)
	  { 
	    fprintf (stderr,"AMBIGUOUS QUALIFIER %s\n", *argv);
	    exit (1);
	  }
       }
      else if (filename == 0)
	 filename = *argv;
      else
       { 
	 fprintf (stderr,"ILLEGAL PARAMETER %s\n", *argv);
	 exit (1);
       }
    }

  /* make decisons about the parameters */

  if (filename == 0)
    { fprintf (stderr,"ERROR: No input file specified\n");
      exit (1);
    }
   if (start == -1)
    { start = DEFAULT_START;
    }
   if (entry == -1)
    { entry = start;
    }

  signal(SIGALRM,timeout_handler);
		  
  /* open the file descriptors for the terminal and lsi */

  tty_fd = open ("/dev/tty", READ_WRITE);
  if (tty_fd < 0)
   {
    fprintf (stderr, "Error opening terminal channel \n");
    exit (1);
   }
  line_fd = open (pdp_device, READ_WRITE);
  if (line_fd < 0)
   {
    fprintf (stderr, "Error opening lsi channel: %s\n", pdp_device);
    exit (1);
   }

#if TURN_OFF_SUPPORTED
  if (turn_off)
   { old_active = set_line_active(pdp_device,0);
   }
#endif

  if (getenv("RCCL_LIB_GENERIC"))
   { sprintf (boot_file, "%s/moper/lsiboot", getenv("RCCL_LIB_GENERIC"));
   }
  else
   { fprintf (stderr,"Error: environment variable RCCL_LIB_GENERIC not set\n");
     exit (1);
   }

  bootstrap = fopen(boot_file,"r");
  if (bootstrap == NULL)
    { fprintf(stderr,"Cannot find bootstrap file '%s'\n",boot_file);
      maybe_turnoff();
      exit (1);
    }

  /* Get the filename for the file to be downloaded, and open it */
  
  load_fd = open (filename, READ);
  if (load_fd < 0)
   { 
     fprintf (stderr, "Error opening load file %s\n",filename);
     maybe_turnoff();
     exit (1);
   }


  /* Assign appropriate device characteristics to the lsi channel */

  termlineGet (line_fd, &old_linestate);
  new_linestate = old_linestate;
  termlineRawMode (&new_linestate);
  termlineEchoOff (&new_linestate);
  termlineSet (line_fd, &new_linestate);

  /* try and determine whether to request a reset */

#define IBUF_SIZE 1024
   { static char iBuf[IBUF_SIZE];
     int numChars, last;
     int inODT;

     inODT = 0;

     put_lsi ("?");
     numChars = read_lsi (iBuf, IBUF_SIZE, 1.0); /* read for 1.0 second */
     if (numChars == 5)
      { if (strncmp (iBuf, "??\r\n@", 5) == 0)
	 { inODT = 1;
	 }
      }

     if (!inODT)
      { printf ("** Power up, or reset, the LSI-11\n");

	get_lsi_char();
        if (wait_lsi ("(Y/N)? ", "@", 0) == 1)
	 { send_lsi ("Y\r");
           if (wait_lsi ("(Y/N)? ", "@", 0) == 1)
	    { send_lsi ("Y\r");
	      if (wait_lsi ("(Y/N)? ", ".", 0) == 1) /* If controller asks */
	       { send_lsi ("Y\r");		     /* about calibration */
		 wait_lsi ("not loaded", 0);	     /* file */
		 wait_lsi (".", 0);
	       }
	      send_lsi ("DONE\r");
	      wait_lsi ("(Y/N)? ", 0);
	      send_lsi ("Y\r");
	      wait_lsi ("@", 0);
	    }
	 }
      }
   }

  load_value ("4", "6");		/* set up to halt on address error */
  load_value ("6", "0");

  load_value ("250", "252");
  load_value ("252", "0");

  load_value ("114", "116");
  load_value ("116", "0");

  load_value ("14", "16");
  load_value ("16", "0");

  load_value ("24", "26");
  load_value ("26", "0");

  load_value ("244", "246");
  load_value ("246", "0");

  load_value ("100", "102");		/* Send down something to disable */
  load_value ("102", "2");		/* clock interrupts and the like */

  if (cpuType == LSI11_73)
   { load_value ("17777572", "0");		/* clear memory mapping */
   }

  load_value ("RS", "340");

  sprintf (octal_num, "%06o", BOOTSTART);
  send_lsi (octal_num);
  send_lsi ("/");
  wait_lsi (" ", 0);

   while (fscanf(bootstrap,"%o",&i) == 1)
    { sprintf(octal_num,"%06o",i);
      send_lsi (octal_num);
      send_lsi ("\n");
      wait_lsi (" ", 0);
    }
   fclose(bootstrap);
  
  send_lsi ("\r");
  wait_lsi ("@", 0);
  send_lsi ("R7/");
  wait_lsi (" ", 0);
  sprintf (octal_num, "%06o", BOOTSTART);
  send_lsi (octal_num);
  send_lsi ("\r");
  wait_lsi ("@", 0);

  send_lsi ("P");

  /* now read in the information */

  read (load_fd, &loadhead, HEADER_SIZE);

  if (bigEndianHost)
    { loadhead.a_text = swapShort (loadhead.a_text);
      loadhead.a_data = swapShort (loadhead.a_data);
      loadhead.a_bss = swapShort (loadhead.a_bss);
    }

  if (!silent)
   { printf (" a.text %d, a.data %d, a.bss %d\n",
	     loadhead.a_text, loadhead.a_data, loadhead.a_bss);
   }
  
  num_bytes = loadhead.a_text + loadhead.a_data /* + loadhead.a_bss*/;

  last_packet_size = num_bytes % BUFFER_SIZE;
  num_packets = num_bytes / BUFFER_SIZE;
  if (last_packet_size > 0)
     ++num_packets;
  else
     last_packet_size = BUFFER_SIZE;

  if (!silent)
    { printf ("\n");
      printf ("Starting address is %d\n", start);
      printf ("Entry address is %d\n", entry);
      printf ("Loading %d bytes ...\n", num_bytes);
    }

   debug = verbose;
   verbose = 0;		/* no verbose in during binary transmission */
  
  /* Now down load the starting address, number of bytes, the data, and 
   * the entry point */

  writeShort (line_fd, start);
  writeShort (line_fd, num_packets);

  packet_count = 1;
  while (packet_count <= num_packets)
   { unsigned int k;
     if (packet_count == num_packets)
	loadbuf.nbytes = last_packet_size;
     else
        loadbuf.nbytes = BUFFER_SIZE;
     read (load_fd, loadbuf.buffer, loadbuf.nbytes);

     checksum = 0;
     for (k=0; k<loadbuf.nbytes; k++)
      { checksum += loadbuf.buffer[k];
      }

     num_trys = 1;
     lsi_char = 'R';
     while (lsi_char != 'A' && num_trys <= MAXTRYS)
      { 
	if (num_trys > 1)
	 { if (lsi_char == 'R')
	    { fprintf (stderr,
"Checksum error: retransmitting packet %d\n", packet_count);
	    }
	   else
	    { fprintf (stderr, "Communication error with LSI-11, packet %d\n",
			       packet_count);
	      printf ("lsi_char = %d\n", lsi_char);
	      maybe_turnoff();
	      exit (-1);
	    }
	 }
	writeShort (line_fd, loadbuf.nbytes);
	write (line_fd, loadbuf.buffer, loadbuf.nbytes);
	writeShort (line_fd, checksum);
	lsi_char = get_lsi_char();
	num_trys++;
      }
     if (lsi_char != 'A')
      { fprintf (stderr, "ERROR transmitting data to LSI\n");
	exit (1);
      }
     if (debug)
      { fprintf (stderr, "packet %d of %d\n", packet_count, num_packets);
      }
     packet_count++;
   }

  if (!silent)
   { printf ("\nLoading completed.\n");
   }

  if (terminate)
    { 
      writeShort (line_fd, entry);
      writeShort (line_fd, loadhead.a_bss);
      if (!silent)
         printf ("Program terminated\n");
      maybe_turnoff();
      exit (0);
    }

  /* Now fork into two processes. The parent will take characters from
   * the terminal and send them on to the lsi; the child will take 
   * characters from the lsi and send them on to the terminal. A trap 
   * for ^C is enabled that can can the process */

  if ((pid = fork()) == 0)
   { 
     from_fd = line_fd;
     to_fd = tty_fd;
   }
  else
   {
     from_fd = tty_fd;
     to_fd = line_fd;
     writeShort (line_fd, entry);
     writeShort (line_fd, loadhead.a_bss);


     /* Now enter into passall: */
     /* Assign appropriate device characteristics to the term channel */

     termlineGet (tty_fd, &old_ttystate);
     new_ttystate = old_ttystate;
     termlineRawMode (&new_ttystate);
     termlineEchoOff (&new_ttystate);
     termlineSet (tty_fd, &new_ttystate);
   }

  while (!terminate)
   { 
     num_chars = read (from_fd, buffer, BUF_SIZE);
     
     if (num_chars <= 0)
        escape ("Error performing read\n");

     for (i=0; i<num_chars; i++)
        if ((buffer[i] == INTERRUPT_CHAR) && (pid != 0))
	   break;

     write (to_fd, buffer, i);
     
     if (i != num_chars)
        terminate = 1;
   }
  escape ("Program exit\n");
 }

/* Private copy of testinput(), so we don't rely on the version in JL's
   library */

#include <sys/time.h>
#include <sys/types.h>

#define NUMFDS 32

testinput (fd, time)
int fd;
float time;
{
	struct timeval timeout;		/* timeout variable for select()   */
	fd_set readmask;		/* IO detection mask for select()  */
	int status;			/* status returned by select()	   */
 
	/* turn 'time' into the appropriate entry in struct timeval */
 
	if (time > 2.14748e9)
	 { time = 2.14748e9;
	 }
	if (time < 0)
	 { time = 0;
	 }
	timeout.tv_sec = time;
	timeout.tv_usec = (time - timeout.tv_sec) * 1000000;

	/* Now perform the test on the stdin stream */
 
	FD_ZERO(&readmask);
	FD_SET(fd, &readmask);
	status = select (NUMFDS, &readmask, NULL, NULL, &timeout);
 
	if (status > 1 || status < 0)
	 { return (-1);
	 }
	else
	 { return (status);
	 }
}


int iswhite(c)
char c;
{
 return((c==' ')||(c=='\t'));
}

int matches(s,pat)
register char *s;
register char *pat;
{
 for (;*pat;s++,pat++)
  { if (*s == *pat)
     { continue;
     }
    if (iswhite(*pat) && iswhite(*s))
     { continue;
     }
    return(0);
  }
 return (1);
}

int skipfields(n,str)
int n;
char *str;
{
 int quoted;
 char *cp;

 cp = str;
 quoted = 0;
 while (isspace(*cp))
  { cp ++;
  }
 for (;n>0;n--)
  { while (quoted || !isspace(*cp))
     { quoted ^= (*cp == '"');
       cp ++;
     }
    while (isspace(*cp))
     { cp ++;
     }
  }
 return(cp-str);
}

#if TURN_OFF_SUPPORTED
int set_line_active(tty,active)
char *tty;
int active;
{
 static char utmp_file[] = "/etc/utmp";
 static char ttys_file[] = "/etc/ttys";
 char ttys_line[1024];
 struct utmp ut;
 char *dev;
 int fd;
 FILE *f;
 int pos;
 int old;
 int didchange;
 int i;

 if (strncmp(tty,"/dev/",5) != 0)
  { return(0);
  }
 if (geteuid() != 0)
  { return(0);
  }
 dev = tty + 5;
 fd = open(utmp_file,O_RDONLY,0);
 if (fd < 0)
  { fprintf(stderr,"%s: cannot open %s: ",argvec[0],utmp_file);
    perror(0);
    exit(1);
  }
 while (read(fd,&ut,sizeof(struct utmp)) == sizeof(struct utmp)) /* foo */
  { if (strncmp(ut.ut_line,dev,sizeof(ut.ut_line)) == 0)
     { if (ut.ut_name[0])
	{ fprintf(stderr,"%s: %s is busy\n",argvec[0],tty);
	  exit(1);
	}
       break;
     }
  }
 close(fd);
 f = fopen(ttys_file,"r+");
 if (f == NULL)
  { return(0);
  }
 old = 0;
 didchange = 0;
 while (1)
  { pos = ftell(f);
    if (fgets(ttys_line,sizeof(ttys_line),f) != ttys_line)
     { break;
     }
    if (strncmp(ttys_line,dev,strlen(dev)) == 0)
     { i = skipfields(3,ttys_line);
       if (i > 0)
	{ int ok;
	  ok = 0;
	  if (matches(&ttys_line[i],"off "))
	   { old = 0;
	     ok = 1;
	   }
	  if (matches(&ttys_line[i],"on  "))
	   { old = 1;
	     ok = 1;
	   }
	  if (ok && (old != active))
	   { if (active)
	      { ttys_line[i+1] = 'n';
		ttys_line[i+2] = ' ';
	      }
	     else
	      { ttys_line[i+1] = 'f';
		ttys_line[i+2] = 'f';
	      }
	     fseek(f,pos,0);
	     fputs(ttys_line,f);
	     didchange = 1;
	   }
	}
       break;
     }
  }
 fclose(f);
 if (didchange)
  { kill(1,SIGHUP);
    sleep(1);
  }
/* printf ("%s logins\n", active ? "enabled" : "disabled"); */
 return(old);
}
#endif

maybe_turnoff()
{
#if TURN_OFF_SUPPORTED
 if (old_active >= 0)
  { set_line_active(pdp_device,old_active);
  }
#endif
}
