/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*---------------------------------------------------------------------------
 *		
 * MODULE NAME:	signalRef
 *
 * DESCRIPTION:	These routines connect and disconnect RCI tasks from
 *		UNIX signals. The code is analogous to that used
 *		in 'interruptRef.c'.
 *
 * AUTHOR/DATE:	John Lloyd, Fall 1987
 *
 * MODIFICATION HISTORY:
 *
 *--------------------------------------------------------------------------*/

#include <errno.h>
#include "../h/signalFxns.h"

#include "../h/cdefs.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/rciTask.h"

#include "../rci/monitorDefs.h"
#include "../rci/interruptDev.h"
#include "../rci/link.h"

#ifdef HAS_POSIX_SIGNALS
#define sigvec sigaction
#define sv_mask sa_mask
#define sv_handler sa_handler
#define sv_flags sa_flags
#endif

static struct sigvec svec;
static struct sigvec oldvecs[NSIG];
static INTERDEV *signalList[NSIG];

int rciSignalRefCount = 0;

/* ----- Error Messages. Defined externally to save space ----- */

static char *badSignalMsg	= "%s -- %d is not a valid signal\n";
static char *reservedSignalMsg	= "%s -- signal %d reserved is to RCI\n";
static char *outOfMemoryMsg	= "%s -- out of memory\n";
static char *outOfTaskEntriesMsg = "%s -- out of task entries\n";
static char *badSigvecMsg	= "%s -- sigvec() failed, errno %d\n";

#ifdef Qnx_TARGET
#pragma aux setup_es = "mov ax,es","push ds","pop es"
extern unsigned setup_es (void);
#pragma aux restore_es = "mov es,ax" parm [eax]
extern void restore_es ();
#endif

static SIGHAN_RETURN signalHandler(signo)
int signo;
{
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	/* ARGH!! Solaris doesn't allow SIGALRM to be
			   blocked (in multi-threaded processes), so we 
			   turn SIGALRM into SIGVTALRM, which can be 
			   blocked */
	if (signo == SIGALRM)
	 { kill (getpid(), SIGVTALRM);
	   return;
	 }
	else if (signo == SIGVTALRM)
	 { signo = SIGALRM;
	 }
#endif
#ifdef Qnx_TARGET
        unsigned save_es = setup_es();
#endif
	monitorInterruptRun (signalList[signo]->queue);
#ifdef Qnx_TARGET
        restore_es (save_es);
#endif
}

static doCleanup ()
{
	INTERDEV *sig;
	int k;

	for (k=0; k<NSIG; k++)
	 { if (sigsetHasSig (svec.sv_mask, k))
	    { sigvec (k, &oldvecs[k], NULL);
	      sig = signalList[k];
	      signalList[k] = NULL;
	      if (sig)
	       { FREE (sig);
	       }
	    }
	 }
	svec.sv_mask = sigsetClear();
	rciSignalRefCount = 0;
}

static initDone = 0;
#define INIT_CHECK()	if (!initDone) doInit();
static doInit()
{
	/* Called once per program to initialize data structures */

	initDone = 1;
	svec.sv_handler = signalHandler;
	svec.sv_mask = sigsetClear();
	svec.sv_flags = 0;
#ifdef Vxw_TARGET
	requestExitFxn(doCleanup,0);
	requestFlagInit(&initDone);
#endif
}

referenceSignal (td, signo)
RCI_DESC *td;
int signo;
{	
	/* Connect the task 'td' to the UNIX signal 'signo'.
	   Place the task on the list of tasks for the RCI monitor 
	   to poll when this signal is delivered.

	   RCI takes over the named signal the first time
	   a task is attached to it. RCI restores the signal when
	   the last task drops the reference.
	 */

	char *fxnName = "referenceSignal()";
	INTERDEV *sig;
	
	INIT_CHECK();

	/* check that signo is a valid signal */

	if (signo <= 0 || signo > (NSIG-1))
  	 { errorMsg (EBadDev, badSignalMsg, fxnName, signo);
	   return (-1);
	 }
	if (signo == SIGHUP)
	 { errorMsg (EBadDev, reservedSignalMsg, fxnName, signo);
	   return (-1);
	 }
#ifdef SIGXCPU
	if (signo == SIGXCPU)
	 { errorMsg (EBadDev, reservedSignalMsg, fxnName, signo);
	   return (-1);
	 }
#endif
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	if (signo == SIGVTALRM)
	 { errorMsg (EBadDev, reservedSignalMsg, fxnName, signo);
	   return (-1);
	 }
#endif

	/* look to see if we are maintaining a descriptor for the signal */

	if (!(sigsetHasSig(svec.sv_mask, signo)))
	 { 
	   /* If we aren't maintaining a descriptor, set one up */
	   int k;

	   if ((sig = ALLOC (INTERDEV)) == NULL)
	    { errorMsg (EMemory, outOfMemoryMsg, fxnName);
	      return (-1);
	    }
	   if ((sig->queue = getTaskEntry()) == NULL)
	    { errorMsg (EInternal, outOfTaskEntriesMsg, fxnName);
	      return (-1);
	    }
	   sig->queue->next = sig->queue->prev = sig->queue;
	   sig->devnum = signo;

	   signalList[signo] = sig;
	   svec.sv_mask = sigsetAdd (svec.sv_mask, signo);
printf ("calling sigvec, signo=%d\n", signo);
	   if (sigvec (signo, &svec, &oldvecs[signo]) < 0)
	    { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
	      return (-1); 
	    }
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	   if (signo == SIGALRM)
	    { /* Take over SIGVTALRM as well. See comment at the top
		 of signalHandler() */
	      svec.sv_mask = sigsetAdd (svec.sv_mask, SIGVTALRM);
	      if (sigvec (SIGVTALRM, &svec, &oldvecs[SIGVTALRM]) < 0)
	       { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
		 return (-1); 
	       }
	    }
#endif
printf ("done\n");
	   for (k=1; k<NSIG; k++)
	    { 
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	      if ((signo != SIGALRM && k != signo) ||
		  (signo == SIGALRM && k != signo && k != SIGVTALRM))
#else
	      if (k != signo)
#endif
	       { if (sigsetHasSig (svec.sv_mask, k))
		  { 
printf (" * calling sigvec, signo=%d\n", k);
		    if (sigvec (k, &svec, NULL) < 0 && errno != EINVAL)
		     { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
		       return (-1); 
		     }
printf (" * done\n", k);
		  }
	       }
	    }
	 }
	else
	 { sig = signalList[signo];
	 }
	
	/* Make sure the task is referenced in the queue for this signal */

	if (insertTaskEntry(td, sig->queue) < 0)
	 { errorMsg (EInternal, outOfTaskEntriesMsg, fxnName);
	   return (-1);
	 }

	rciSignalRefCount++;
	return (1);
}

dereferenceSignal (td, signo)
RCI_DESC *td;
int signo;
{
	/* Remove 'td' from the list of tasks attached to signal 'signo'. */

	char *fxnName = "dereferenceSignal()";
	INTERDEV *sig;
	
	INIT_CHECK();

	/* find the device signo */

	if (!(sigsetHasSig (svec.sv_mask, signo)))
	 { return (0);
	 }
	sig = signalList[signo];

	deleteTaskEntry (td, sig->queue);

	if (sig->queue->next == sig->queue)
	 { int k;
	   if (sigvec (signo, &oldvecs[signo], NULL) < 0)
	    { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
	      return (-1);
	    }
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	   if (signo == SIGALRM)
	    { if (sigvec (SIGVTALRM, &oldvecs[SIGVTALRM], NULL) < 0)
	       { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
		 return (-1);
	       }
	    }
#endif
	   freeTaskEntry (sig->queue);
	   
	   signalList[signo] = NULL;
	   svec.sv_mask = sigsetDel (svec.sv_mask, signo);
#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
	   if (signo == SIGALRM)
	    { svec.sv_mask = sigsetDel (svec.sv_mask, SIGVTALRM);
	    }
#endif
	   FREE (sig);
	
	   for (k=0; k<NSIG; k++)
	    { if (sigsetHasSig (svec.sv_mask, signo))
	       { if (sigvec (k, &svec, NULL) < 0)
		  { errorMsg (EInternal, badSigvecMsg, fxnName, errno);
		    return (-1);
		  }
	       }
	    }
	 }
	rciSignalRefCount--;
	return (1);
}

#define MASK_STK_SIZE	32

static SigSet maskStk[MASK_STK_SIZE];
static int maskPtr = 0;


blockRciSignals()
{
	/* Prevent delivery of signals to which RCI tasks are connected */

	SigSet oldmask;

	INIT_CHECK();

	oldmask = signalMaskAdd(svec.sv_mask);
	maskStk[maskPtr++] = oldmask;
}

unblockRciSignals()
{
	/* Re-enable delivery of task signals */

	INIT_CHECK();

	signalMaskSet (maskStk[--maskPtr]);
}

dereferenceAllSignals ()
{
	/* called by rciClose() */

	INTERDEV *sig;
	int k;

	INIT_CHECK();

	for (k=0; k<NSIG; k++)
	 { if (sigsetHasSig (svec.sv_mask, k))
	    { TASK_ENTRY *entry, *entry_next;

	      sig = signalList[k];
	      if (sig)
	       { for (entry=sig->queue->next; entry!=sig->queue; 
					      entry=entry_next)
		  { entry_next = entry->next;
		    freeTaskEntry (entry);
		  }
	       }
	      sigvec (k, &oldvecs[k], NULL);
	      if (sig)
	       { freeTaskEntry (sig->queue);
		 FREE (sig);
	       }
	      signalList[k] = NULL;
	    }
	 }
	svec.sv_mask = sigsetClear();
	rciSignalRefCount = 0;
}
