/* Tgen API (Application Programming interface).

   The routines in this file provide the programming interface
   to the TGEN package. */

#include <stdio.h>
#include <robotMath.h>
#include <robot_kynvar.h>
#include <errorStack.h>
#include "tgen.h"
#include <vector.h>

Tgen *tgenAlloc (robotName)
char *robotName;
{
	Tgen *tg;
	MotionRequest *req0, *req1;
	MotionControl *curr;
	int j;

	tg = (Tgen*)calloc(1, sizeof(Tgen));

	/* motion request queue is initially clear */
	tg->reqHead = NULL;
	tg->reqTail = NULL;

	/* allocate four motion control blocks */
	tg->prevMot =     (MotionControl*)calloc(1, sizeof(MotionControl));
	tg->currMot =     (MotionControl*)calloc(1, sizeof(MotionControl));
	tg->nextMot =     (MotionControl*)calloc(1, sizeof(MotionControl));
	tg->nextnextMot = (MotionControl*)calloc(1, sizeof(MotionControl));

	/* allocate two motion request blocks and initialize the motion
	   queue with them. */

	req0 = (MotionRequest*)calloc(1, sizeof(MotionRequest));
	req1 = (MotionRequest*)calloc(1, sizeof(MotionRequest));
	
	req0->prev = NULL;
	req0->next = req1;
	req1->prev = req0;
	req1->next = NULL;
	tg->queueTail = req1;

	tg->prevMot->req = req0;
	tg->currMot->req = req1;
	tg->nextMot->req = NULL;
	tg->nextnextMot->req = NULL;

	/* Now we must initialize the contents of the two request blocks
	   and the four motion control blcoks. The "prev" control block
	   can be ignored since we have already "transitioned" to the
	   initial idle motion. Correspondingly, the first request block
	   can be ignored, except for setting its id to 0. The "next" and 
	   "nextnext" control blocks can be ignored because they have 
	   no motions associated with them. That means we need only be 
	   concerned with the "curr" control block and its associated 
	   motion request */

	req0->id = 0;

	curr = tg->currMot;

	curr->transCnt = 0;
	curr->segCnt = 1;
	curr->cnt = 0;
	curr->scale = 1.0;
	curr->status = TGEN_ACTIVE;

	/* motion velocities and displacements are all zero, regardless ... */

	zeroVf (curr->jvel.v, NJNTS);
	zeroVf (curr->delJvel.v, NJNTS);
	zeroVf (curr->prevJvel.v, NJNTS);
	zeroDiff (&curr->cvel);
	zeroDiff (&curr->delCvel);
	zeroDiff (&curr->prevCvel);

	/* max velocity and accel specs for req1 can be ignored. 
	   Also, so can the time specs. */

	zeroDiff (&req1->cdrive);
	zeroVf (req1->jdrive.v, NJNTS);
	req1->flags = TGEN_STOP | TGEN_IDLE;
	req1->id = 0;
	identTrsf(&req1->TL);

	/* Still to be initialized: 
	   req1->pathType, req1->targType, req1->TC, req1->conf, req1->jc */
	
	/* Now finish initializing the Tgen structure: */

	tg->timeCnt = 0;
	tg->flags = 0;
	tg->alterFxn = NULLF;
	tg->alterArg = 0;
	tg->alterPtr = 0;

	if (robotName)
	 { if ((tg->jls = getJls(robotName)) == NULL)
	    { errorMsg (TGNoRobotJls, "tgenAlloc(): can't get JLS for %s\n",
				       robotName);
	      return (NULL);
	    }
	  if ((tg->kyn = getKyn(robotName, (int*)0)) == NULL)
	    { errorMsg (TGNoRobotKyn, "tgenAlloc(): can't get KYN for %s\n",
				       robotName);
	      return (NULL);
	    }
	  tg->name = robotName;
	  tg->flags |= TGEN_ROBOT_ATTACHED;
	 }
	tg->diagHead = NULL;
	tg->diagTail = NULL;

	/* Still to be initialized or changed: 
	   tg->j6, tg->T6, tg->conf, tg->flags, tg->reqCtrl.movecPathType, 
	   tg->reqCtrl.confReq */

	/* Now allocate the motion request template */

#if 0
	tg->reqTmpl = (MotionRequest*)calloc(1,sizeof(MotionRequest));
#endif

	tg->reqTmpl.maxTransVel = 500;
	tg->reqTmpl.maxTransAcc = 5000;
	tg->reqTmpl.maxRotVel = 180*DEGTORAD;
	tg->reqTmpl.maxRotAcc = 360*4*DEGTORAD;
	for (j=0; j<NJNTS; j++)
	 { if (robotName)
	    { tg->reqTmpl.maxJntVel.v[j] = tg->jls->nominalVel[j];
	      tg->reqTmpl.maxJntAcc.v[j] = tg->jls->nominalAcc[j];
	    }
	   else
	    { tg->reqTmpl.maxJntVel.v[j] = 180*DEGTORAD;
	      tg->reqTmpl.maxJntAcc.v[j] = 360*4*DEGTORAD;
	    }
	 }
	tg->reqTmpl.segTime = -1;
	tg->reqTmpl.transTime = -1;
	tg->reqTmpl.flags = TGEN_STOP | TGEN_IDLE;
	tg->reqTmpl.id = 0;
	identTrsf (&tg->reqTmpl.TL);

	zeroDiff (&tg->reqTmpl.cdrive);
	zeroVf (tg->reqTmpl.jdrive.v, NJNTS);
	
	/* Still to be initialized or changed: 
	   reqTmpl.pathType */

	return (tg);
}

tgenInitGen (tg, interval, T6, conf, j6, targType)
Tgen *tg;
double interval;
TRSF *T6;
unsigned long conf;
JntVec *j6;
int targType;
{
	TRSF T6init;
	unsigned long confInit;
	JntVec j6init;
	MotionRequest *req, *reqNext;
	MotionControl *curr;

	if (targType == TGEN_JNT_TARG)
	 { if (initCalcJnt (tg, &T6init, &confInit, &j6init, j6) < 0)
	    { return (-1);
	    }
	 }
	else
	 { if (initCalcCart (tg, &T6init, &confInit, &j6init, T6, conf) < 0)
	    { return (-1);
	    }
	 }

	for (req=tg->reqHead; req; req=reqNext)
	 { reqNext = req->next;
	   free (req);
	 }
	for (req=tg->nextMot->req; req; req=reqNext)
	 { reqNext = req->next;
	   free (req);
	 }
	
	tg->sampInterval = interval;

	tg->reqHead = NULL;
	tg->reqTail = NULL;
	tg->queueTail = tg->currMot->req;

	tg->nextMot->req = NULL;
	tg->nextnextMot->req = NULL;

	tg->timeCnt = 0;
	
	tgenClearDiag (tg);

	tg->reqTmpl.segTime = -1;
	tg->reqTmpl.transTime = -1;
	tg->reqTmpl.flags = TGEN_STOP | TGEN_IDLE;

	curr = tg->currMot;

	if (targType == TGEN_JNT_TARG)
	 { curr->req->pathType = TGEN_JNT_PATH;
	   curr->req->targType = TGEN_JNT_TARG;
	   tg->reqTmpl.pathType = TGEN_JNT_PATH;
	   tg->reqCtrl.movecPathType = TGEN_JNT_PATH;
	 }
	else
	 { curr->req->pathType = TGEN_CART_PATH;
	   curr->req->targType = TGEN_CART_TARG;
	   tg->reqTmpl.pathType = TGEN_CART_PATH;
	   tg->reqCtrl.movecPathType = TGEN_CART_PATH;
	 }
	
	curr->req->TC = T6init;
	curr->req->TB = T6init;
	curr->req->conf = confInit;
	curr->req->jc = j6init;
	tg->T6 = T6init;
	tg->conf = confInit;
	tg->j6 = j6init;
	identTrsf (&tg->Alter);
	tg->oldJ6 = j6init;
	tg->reqTmpl.TC = T6init;
	tg->reqTmpl.TB = T6init;
	tg->reqTmpl.conf = confInit;
	tg->reqTmpl.jc = j6init;
	tg->reqCtrl.T6C = T6init;
	tg->reqCtrl.conf = confInit;
	tg->reqCtrl.jc = j6init;
	tg->reqCtrl.confReq = confInit;
	return (1);
}

tgenInitJnt (tg, interval, j6)
Tgen *tg;
double interval;
JntVec *j6;
{
	int code;

	code = tgenInitGen (tg, interval, (TRSF*)NULL, 0, j6, TGEN_JNT_TARG);
	return (code);
}

tgenInitCart (tg, interval, T6, conf)
Tgen *tg;
double interval;
TRSF *T6;
unsigned long conf;
{
	int code;

	code = tgenInitGen (tg, interval, T6, conf,(JntVec*)0,TGEN_CART_TARG);
	return (code);
}

initCalcCart (tg, T6, confp, j6, T6in, confIn)
Tgen *tg;
TRSF *T6;
unsigned long *confp;
JntVec *j6;
TRSF *T6in;
unsigned long confIn;
{
	int status;
	extern char* jointDescStr();
	char buf[32];

	if (T6in == NULL)
	 { if (tg->flags & TGEN_ROBOT_ATTACHED)
	    { copyVf (j6->v, tg->jls->bestAngles, NJNTS);
	      fwdKinematics (T6, confp, j6->v, tg->kyn, (void*)0);
	    }
	   else
	    { identTrsf (T6);
	      *confp = 0;
	    }
	 }
	else
	 { *T6 = *T6in;
	   *confp = confIn;
	 }

	if (tg->flags & TGEN_ROBOT_ATTACHED)
	 { if (T6in)
	    { status = invKinematics (j6->v, T6, *confp, 
		   		      (float*)0, tg->kyn, tg->jls, (void*)0);
	      if (status & KYN_JOINT_LIMIT)
	       { errorMsg (TGJntLimit,
"tgenInitCart(): joint limit exceeded for joints %s\n", 
			  jointDescStr(buf, status & KYN_JOINT_LIMIT));
		 return (-1);
	       }
	      else if (status & KYN_CANT_REACH)
	       { errorMsg (TGWorkspaceLimit, 
"tgenInitCart(): position out of reach\n");
		 return (-1);
	       }
	      else if (status & KYN_SINGULAR && 
		       tg->reqTmpl.pathType != TGEN_JNT_PATH)
	       { errorMsg (TGSingular,
"tgenInitCart(): singular position; can't be reached with Cartesian path\n");
		 return (-1);
	       }
	    }
	 }
	else
	 { identTrsf (T6);
	   *confp = 0;
	   zeroVf (j6->v, NJNTS);
	 }
	return (1);
}

initCalcJnt (tg, T6, confp, j6, j6in)
Tgen *tg;
TRSF *T6;
unsigned long *confp;
JntVec *j6, *j6in;
{
	int status;
	
	if (j6in == NULL)
	 { if (tg->flags & TGEN_ROBOT_ATTACHED)
	    { copyVf (j6->v, tg->jls->bestAngles, NJNTS);
	    }
	   else
	    { zeroVf (j6->v, NJNTS);
	    }
	 }
	else
	 { if (tg->flags & TGEN_ROBOT_ATTACHED)
	    { int j;
	      for (j=0; j<NJNTS; j++)
	       { if (j6in->v[j] < tg->jls->jmin[j] ||
		     j6in->v[j] > tg->jls->jmax[j])
		  { errorMsg (TGJntLimit, 
"tgenInitJnt(): joint %d out of bounds at %8.3f\n", j+1, RADTODEG*j6in->v[j]);
		    return (-1);
		  }
	       }
	    }
	   copyVf (j6->v, j6in->v, NJNTS);
	 }
	if (tg->flags & TGEN_ROBOT_ATTACHED)
	 { fwdKinematics (T6, confp, j6->v, tg->kyn, (void*)0);
	 }
	else
	 { identTrsf (T6);
	   *confp = 0;
	 }
	return (1);
}

void tgenFree(tg)
Tgen *tg;
{
	MotionRequest *req, *reqNext;

	for (req=tg->reqHead; req; req=reqNext)
	 { reqNext = req->next;
	   free (req);
	 }
	for (req=tg->prevMot->req; req; req=reqNext)
	 { reqNext = req->next;
	   free (req);
	 }
	
	free (tg->prevMot);
	free (tg->currMot);
	free (tg->nextMot);
	free (tg->nextnextMot);
	free (tg);	
}

tgenSetPath (tg, type)
Tgen *tg;
int type;
{
	int oldType;

	oldType = tg->reqCtrl.movecPathType;
	if (!(tg->flags & TGEN_ROBOT_ATTACHED) && type == TGEN_JNT_PATH)
	 { errorMsg (TGNoJntMotion, "tgenSetPath(): no robot attached\n");
	   return (-1);
	 }
	if (type != TGEN_JNT_PATH && type != TGEN_CART_PATH)
	 { errorMsg (TGBadArg, "tgenSetPath(): illegal path type\n");
	   return (-1);

	 }
	tg->reqCtrl.movecPathType = type;
	return (oldType);
}

tgenSetConf (tg, conf)
Tgen *tg;
unsigned long conf;
{
	tg->reqCtrl.confReq = conf;
	return (1);
}

tgenSetCartVel (tg, tvel, rvel)
Tgen *tg;
double tvel, rvel;
{
	if (tvel == 0 || rvel == 0)
	 { errorMsg (TGBadArg,
"tgenSetCartVel(): velocity values cannot be zero\n");
	   return (-1);
	 }
	if (tvel > 0)
	 { tg->reqTmpl.maxTransVel = tvel;
	 }
	if (rvel > 0)
	 { tg->reqTmpl.maxRotVel = rvel*DEGTORAD;
	 }
	return (1);
}

void tgenGetCartVel (tg, tvelp, rvelp)
Tgen *tg;
double *tvelp, *rvelp;
{
	if (tvelp)
	 { *tvelp = tg->reqTmpl.maxTransVel;
	 }
	if (rvelp)
	 { *rvelp = RADTODEG*tg->reqTmpl.maxRotVel;
	 }
}

tgenSetJntVel (tg, vel, j)
Tgen *tg;
double vel;
int j;
{
	if (vel == 0)
	 { errorMsg (TGBadArg,
"tgenSetJntVel(): velocity values cannot be zero\n");
	   return (-1);
	 }
	if (j >= NJNTS)
	 { errorMsg (TGBadArg, "tgenSetJntVel(): bad index\n");
	   return (-1);
	 }
	if (j >= 0)
	 { tg->reqTmpl.maxJntVel.v[j] = DEGTORAD*vel;
	 }
	else
	 { for (j=0; j<NJNTS; j++)
	    { tg->reqTmpl.maxJntVel.v[j] = DEGTORAD*vel;
	    }
	 }
	return (1);
}

double tgenGetJntVel (tg, j)
Tgen *tg;
int j;
{
	if (j < 0 || j >= NJNTS)
	 { errorMsg (TGBadArg, "tgenGetJntVel(): bad index\n");
	   return (-1);
	 }
	return (RADTODEG*tg->reqTmpl.maxJntVel.v[j]);
}

tgenSetCartAcc (tg, tacc, racc)
Tgen *tg;
double tacc, racc;
{
	if (tacc == 0 || racc == 0)
	 { errorMsg (TGBadArg,
"tgenSetCartAcc(): acceleration values cannot be zero\n");
	   return (-1);
	 }
	if (tacc > 0)
	 { tg->reqTmpl.maxTransAcc = tacc;
	 }
	if (racc > 0)
	 { tg->reqTmpl.maxRotAcc = racc*DEGTORAD;
	 }
	return (1);
}

void tgenGetCartAcc (tg, taccp, raccp)
Tgen *tg;
double *taccp, *raccp;
{
	if (taccp)
	 { *taccp = tg->reqTmpl.maxTransAcc;
	 }
	if (raccp)
	 { *raccp = RADTODEG*tg->reqTmpl.maxRotAcc;
	 }
}

tgenSetJntAcc (tg, acc, j)
Tgen *tg;
double acc;
int j;
{
	if (acc == 0)
	 { errorMsg (TGBadArg,
"tgenSetJntAcc(): acceleration values cannot be zero\n");
	   return (-1);
	 }
	if (j >= NJNTS)
	 { errorMsg (TGBadArg, "tgenSetJntAcc(): bad index\n");
	   return (-1);
	 }
	if (j >= 0)
	 { tg->reqTmpl.maxJntAcc.v[j] = DEGTORAD*acc;
	 }
	else
	 { for (j=0; j<NJNTS; j++)
	    { tg->reqTmpl.maxJntAcc.v[j] = DEGTORAD*acc;
	    }
	 }
	return (1);
}

double tgenGetJntAcc (tg, j)
Tgen *tg;
int j;
{
	if (j < 0 || j >= NJNTS)
	 { errorMsg (TGBadArg, "tgenGetJntAcc(): bad index\n");
	   return (-1);
	 }
	return (RADTODEG*tg->reqTmpl.maxJntAcc.v[j]);
}

tgenSetMotionTime (tg, seg, trans)
Tgen *tg;
double seg, trans;
{
	if (seg >= 0)
	 { tg->reqTmpl.segTime = seg;
	 }
	if (trans >= 0)
	 { tg->reqTmpl.transTime = trans;
	 }
	return (1);
}

tgenMoveC (tg, TCd, TL)
Tgen *tg;
TRSF *TCd;
TRSF *TL;
{
	int status;
	extern char* jointDescStr();
	TRSF TC;
	TRSF TD;
	TRSF TB;
	TRSF T6;
	unsigned long conf;
	DSPL cdrive;
	char buf[32];

	if (!(tg->flags & TGEN_ROBOT_ATTACHED) &&
	     (tg->reqTmpl.pathType != TGEN_CART_PATH) &&
	    !(tg->reqTmpl.flags & TGEN_STOP))
	 { tgenStop (tg, 0.0);
	 }
	if (tg->reqCtrl.confReq != tg->reqTmpl.conf && 
	    tg->reqCtrl.movecPathType != TGEN_JNT_PATH)
	 { errorMsg (TGNoCartMotion,
"tgenMoveC(): cannot use Cartesian path mode to change configurations\n");
	   return (-1);
	 }

	/* compute drive parameters, and don't actually try to undertake
	   motions which are less than TGEN_MM_EPSILON (translation) or
	   TGEN_ANG_EPSILON (rotation).

	   Drive is computed so that 
						-1
		T6 TL = TC DRIVE  =>  DRIVE = TC   T6 TL

	   where T6 is the T6 target for the last motion. */

	TC = *TCd;
	if (TL)
	 { multTrsf (&TB, &tg->reqCtrl.T6C, TL);
	 }
	else
	 { TB = tg->reqCtrl.T6C;
	 }
	multLiTrsf (&TD, &TC, &TB);
	trsfToDspl (&cdrive, &TD);
	if (normVect(&cdrive.p) < TGEN_MM_EPSILON)
	 { zeroVect(&cdrive.p);
	   TC.p = TB.p;
	 }
	if (ABS(cdrive.a) < TGEN_ANG_EPSILON)
	 { zeroVect(&cdrive.u);
	   cdrive.a = 0;
	   TC.n = TB.n;
	   TC.o = TB.o;
	   TC.a = TB.a;
	 }
	if (TL)
	 { multRiTrsf (&T6, &TC, TL);
	 }
	else
	 { T6 = TC;
	 }

	conf = tg->reqCtrl.confReq;
	if (tg->flags & TGEN_ROBOT_ATTACHED)
	 { JntVec jcNew;

	   status = invKinematics (jcNew.v, &T6, conf, tg->reqCtrl.jc.v, 
				   tg->kyn, tg->jls, (void*)0);
	   if (status & KYN_JOINT_LIMIT)
	    { errorMsg (TGJntLimit,
"tgenMoveC(): joint limit exceeded for joints %s\n", 
		       jointDescStr(buf, status & KYN_JOINT_LIMIT));
	      return (-1);
	    }
	   else if (status & KYN_CANT_REACH)
	    { errorMsg (TGWorkspaceLimit, 
"tgenMoveC(): target out of reach\n");
	      return (-1);
	    }
	   else if (status & KYN_SINGULAR && 
		    tg->reqTmpl.pathType != TGEN_JNT_PATH)
	    { errorMsg (TGSingular,
"tgenMoveC(): target is singular and can't be reached with Cartesian path\n");
	      return (-1);
	    }
	   subtractVf (tg->reqTmpl.jdrive.v,jcNew.v,tg->reqCtrl.jc.v,NJNTS);
	   tg->reqCtrl.jc = jcNew;
	   tg->reqTmpl.jc = jcNew;
	 }
	tg->reqCtrl.T6C = T6;
	tg->reqCtrl.conf = conf;

	tg->reqTmpl.TC = TC;
	tg->reqTmpl.TB = TB;
	tg->reqTmpl.conf = conf;
	tg->reqTmpl.cdrive = cdrive;
	tg->reqTmpl.targType = TGEN_CART_TARG;
	tg->reqTmpl.pathType = tg->reqCtrl.movecPathType;
	tg->reqTmpl.flags = 0;
	if (TL)
	 { tg->reqTmpl.TL = *TL;
	 }
	else
	 { identTrsf (&tg->reqTmpl.TL);
	 }
	return (doMove (tg));
}

tgenMoveJ (tg, jcd)
Tgen *tg;
JntVec *jcd;
{
	JntVec jc;
	JntVec jdrive;
	int j;

	if (!(tg->flags & TGEN_ROBOT_ATTACHED) &&
	     (tg->reqTmpl.pathType != TGEN_JNT_PATH) &&
	    !(tg->reqTmpl.flags & TGEN_STOP))
	 { tgenStop (tg, 0.0);
	 }

	/* compute drive parameters, but don't undertake any motions
	   which are less than TGEN_ANG_EPSILON */

	jc = *jcd;
	subtractVf (jdrive.v, jc.v, tg->reqCtrl.jc.v, NJNTS);
	for (j=0; j<NJNTS; j++)
	 { if (ABS(jdrive.v[j]) < TGEN_ANG_EPSILON)
	    { jdrive.v[j] = 0.0;
	      jc.v[j] = tg->reqCtrl.jc.v[j];
	    }
	 }

	if (tg->flags & TGEN_ROBOT_ATTACHED)
	 { for (j=0; j<NJNTS; j++)
	    { if (jc.v[j] < tg->jls->jmin[j] || jc.v[j] > tg->jls->jmax[j])
	       { errorMsg (TGJntLimit,
"tgenMoveJ(): joint %d out of range\n", j+1);
		 return (-1);
	       }
	    }
	   fwdKinematics (&tg->reqCtrl.T6C, &tg->reqCtrl.conf, 
			   jc.v, tg->kyn, (void*)0);
	   tg->reqCtrl.confReq = tg->reqCtrl.conf;
	 }
	tg->reqCtrl.jc = jc;

	tg->reqTmpl.jc = jc;
	tg->reqTmpl.jdrive = jdrive;
	tg->reqTmpl.targType = TGEN_JNT_TARG;
	tg->reqTmpl.pathType = TGEN_JNT_PATH;
	tg->reqTmpl.flags = 0;
	return (doMove (tg));
}

tgenStop (tg, time)
Tgen *tg;
double time;
{
	if (time < 0)
	 { errorMsg (TGBadArg, "tgenStop(): negative time value\n");
	   return (-1);
	 }
	zeroDspl (&tg->reqTmpl.cdrive);
	zeroVf (tg->reqTmpl.jdrive.v, NJNTS);
	if (tg->reqTmpl.pathType == TGEN_CART_PATH)
	 { tg->reqTmpl.TB = tg->reqTmpl.TC;
	 }
	tg->reqTmpl.segTime = time;
	tg->reqTmpl.flags = TGEN_STOP;
	return (doMove (tg));
}

doMove (tg)
Tgen *tg;
{
	MotionRequest *req;

	req = (MotionRequest*)malloc(sizeof(MotionRequest));
	*req = tg->reqTmpl;

	req->id = ++tg->reqTmpl.id;	/* XXX could wrap around */

	tgenLock (tg);

	req->next = NULL;
	req->prev = tg->reqTail;
	if (tg->reqTail == NULL)
	 { tg->reqHead = req;
	 }
	else
	 { tg->reqTail->next = req;
	 }
	tg->reqTail = req;	

	tgenUnlock (tg);

	/* reset some things to nominal values */

	tg->reqTmpl.segTime = -1;
	tg->reqTmpl.transTime = -1;		

	return (req->id);
}

tgenIsBusy (tg)
Tgen *tg;
{
	if (tg->reqHead != NULL ||
	    tg->nextMot->req != NULL || 
	    tg->currMot->cnt < tg->currMot->transCnt)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

void tgenEnableDiag (tg, set)
Tgen *tg;
int set;
{
	if (set)
	 { tg->flags |= TGEN_DIAG_OUTPUT;
	 }
	else
	 { tg->flags &= ~TGEN_DIAG_OUTPUT;
	 }
}

void tgenClearDiag (tg)
Tgen *tg;
{
	TgenDiag *tgd, *tgdNext;

	for (tgd=tg->diagHead; tgd; tgd=tgdNext)
	 { tgdNext = tgd->next;
	   if (tgd->mctl)
	    { free (tgd->mctl);
	    }
	   free (tgd);
	 }
	tg->diagHead = NULL;
	tg->diagTail = NULL;
}

TgenDiag *tgenGetDiag (tg)
Tgen *tg;
{
	return (tg->diagHead);
}

tgenRun (tg, cnt)
Tgen *tg;
int cnt;
{
	if (cnt > 0)
	 { while (cnt--)
	    { if (trajectoryCycle(tg) < 0)
	       { return (-1);
	       }
	    }
	 }
	else if (cnt < 0)
	 { do
	    { if (trajectoryCycle(tg) < 0)
	       { return (-1);
	       }
	    }
	   while (tgenIsBusy(tg));
	 }
	return (1);
}

tgenMotionStatus (tg, mid, scalep)
Tgen *tg;
int mid;
double *scalep;
{
	int currId;
	int prevId;
	double scale;
	int status;

	if (mid <= 0 || mid > tg->reqTmpl.id)
	 { errorMsg (TGBadArg, "tgenMotionStatus(): invalid motion id\n");
	   return (-1);
	 }

	tgenLock(tg);
	currId = tg->currMot->req->id;
	prevId = tg->prevMot->req->id;
	if (mid == currId)
	 { scale = tg->currMot->scale;
	   status = tg->currMot->status;
	 }
	else if (mid == prevId)
	 { scale = tg->prevMot->scale;
	   status = tg->prevMot->status;
	 }
	tgenUnlock(tg);

	if (mid != prevId && mid != currId)
	 { int refId;

	   refId = (currId == 0) ? prevId : currId;
	   if (mid < refId)
	    { status = TGEN_COMPLETED;
	      scale = 1.0;
	    }
	   else /* mid > refId */
	    { status = TGEN_PENDING;
	      scale = 0.0;
	    }
	 }
	if (scalep)
	 { *scalep = scale;
	 }
	return (status);
}

int (*tgenAlterFxnSet(tg, fxn, arg, ptr))()
Tgen *tg;
int (*fxn)();
int arg;
void *ptr;
{
	int (*old)();

	old = tg->alterFxn;

	tgenLock(tg);
	
	tg->alterFxn = fxn;
	tg->alterArg = arg;
	tg->alterPtr = ptr;

	tgenUnlock(tg);

	return (old);
}

int (*tgenMonitorFxnSet(tg, fxn, arg, ptr))()
Tgen *tg;
int (*fxn)();
int arg;
void *ptr;
{
	int (*old)();

	old = tg->monitorFxn;

	tgenLock(tg);
	
	tg->monitorFxn = fxn;
	tg->monitorArg = arg;
	tg->monitorPtr = ptr;

	tgenUnlock(tg);

	return (old);
}
