#include <math.h>
#include <stdio.h>
#include "tgen.h"
#include <robotTypes.h>
#include <robotMath.h>
#include <vector.h>
#include <errorStack.h>
#include <robot_kynvar.h>
#include <puma_kynvar.h>

#define EPSILON	(1e-6)

void setVect PROTO((VECT*, double, double, double, double));
void setJntVect PROTO((JntVec*, double, double, double, double, 
					double, double, double));
void bail PROTO((int, char*));
void errorTest PROTO((int, int, int));
void clearTest PROTO((int, int));
int sameFloat PROTO((double, double));
void blendVf PROTO((float*,float*,float*,float*,float*,double,double,int));
void blendRot PROTO((TRSF*,TRSF*,VECT*,TRSF*,VECT*,TRSF*,double,double));
int movec PROTO((Tgen*, double, double, double, double, double);)
int movej PROTO((Tgen*, double, double, double));
int checkMoveC PROTO((int, int, int, TRSF*, TRSF*, unsigned long, DIFF*));
int checkMoveT PROTO((int, int, int, TRSF*, TRSF*, TRSF*, 
		      unsigned long, DIFF*));
int checkMoveJ PROTO((int, int, int, JntVec*, JntVec*, JntVec*));

int dbgHook1;
int dbgHook2;

double alphaFxn PROTO((double));
double betaFxn PROTO((double));

void setVect (vr, x, y, z, s)
VECT *vr;
double x, y, z, s;
{
	vr->x = x*s;
	vr->y = y*s;
	vr->z = z*s;
}

void setJntVec (JntVec *jv, double ang1, double ang2, double ang3, 
			    double ang4, double ang5, double ang6, double s)
{
	zeroVf (jv->v, NJNTS);
	jv->v[0] = ang1*s*DEGTORAD;
	jv->v[1] = ang2*s*DEGTORAD;
	jv->v[2] = ang3*s*DEGTORAD;
	jv->v[3] = ang4*s*DEGTORAD;
	jv->v[4] = ang5*s*DEGTORAD;
	jv->v[5] = ang6*s*DEGTORAD;
}

void bail(loc, msg)
int loc;
char *msg;
{
	fprintf (stderr, "%d:: %s\n", loc, msg);
	exit (1);
}

void errorTest (loc, ret, code)
int loc, ret, code;
{
	if (ret >= 0 || getErrorCode(0) != code)
	 { bail (loc, "bad error return");
	 }
	else
	 { clearErrors();
	 }
}

void clearTest (loc, ret)
int loc, ret;
{
	if (ret < 0)
	 { bail (loc, "function returns error");
	 }
}

sameFloat (d1, d2)
double d1, d2;
{
	if (ABS(d1-d2) < EPSILON)
	 { return (1);
	 }
	else if (ABS(d1-d2)/(ABS(d1)+ABS(d2)) < EPSILON)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

sameVf (v1, v2, n)
float *v1;
float *v2;
int n;
{
	int k;

	for (k=0; k<n; k++)
	 { if (!sameFloat(v1[k], v2[k]))
	    { return (0);
	    }
	 }
	return (1);	
}

sameTrsf (T1, T2)
TRSF *T1, *T2;
{
	if (cmpTrsf (T1, T2) < EPSILON)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

/* functions that blend together paths, used for checking the "tgen" blend
   functions. */

void blendVf (vr, v1, vel1, v2, vel2, s, transTime, n)
float *vr, *v1, *vel1;
float *v2, *vel2;
double s, transTime;
int n;
{
	float vb[12];
	double alpha;
	double bfact;

	alpha = alphaFxn(s);
	bfact = transTime*6*betaFxn(s);

	combineVf (vb, -bfact, vel2, bfact, vel1, n);
	combineVf (vr, 1, vb, 1-alpha, v1, n);
	combineVf (vr, 1, vr, alpha, v2, n);
}

void blendRot (TR, T1, vel1, T2, vel2, TP, s, transTime)
TRSF *TR, *T1, *T2, *TP;
VECT *vel1, *vel2;
double s, transTime;
{
	TRSF TA, TB;
	TRSF T1P, TP2;
	VECT u1, u2;
	float ang1, ang2;
	double vel1Mag, vel2Mag;
	double alpha;
	double bfact;

	alpha = alphaFxn(s);
	bfact = transTime*6*betaFxn(s);

	vel1Mag = normVect(vel1);
	vel2Mag = normVect(vel2);
	zeroVect (&TA.p);
	zeroVect (&TB.p);
	multLiTrsf (&T1P, T1, TP);
	multLiTrsf (&TP2, TP, T2);
	trsfToRot (&u1, &ang1, &T1P);
	trsfToRot (&u2, &ang2, &TP2);
	rotToTrsf (&TA, &u1, alpha*ang1 + RADTODEG*bfact*vel1Mag);
	rotToTrsf (&TB, &u2, alpha*ang2 - RADTODEG*bfact*vel2Mag);
	multTrsf (TR, T1, &TA);
	multTrsf (TR, TR, &TB);

	if (dbgHook1)
	 { printTrsf ("TA> %10.5m\n", &TA);
	   printTrsf ("TB> %10.5m\n", &TB);
	 }
}

/* functions for requesting simple test motions */

movec (tg, x, y, roll, pitch, yaw)
Tgen *tg;
double x, y, roll, pitch, yaw;
{
	TRSF TC;

	xyzToTrsf (&TC, x, y, 0.0);
	rpyToTrsf (&TC, roll, pitch, yaw);
	return (tgenMoveC (tg, &TC, (TRSF*)0));
}

movej (tg, ang1, ang2, ang3)
Tgen *tg;
double ang1, ang2, ang3;
{
	JntVec jc;

	zeroVf (jc.v, NJNTS);
	jc.v[0] = ang1*DEGTORAD;
	jc.v[1] = ang2*DEGTORAD;
	jc.v[2] = ang3*DEGTORAD;
	return (tgenMoveJ (tg, &jc));
}

/* functions for checking moitons */

static TgenDiag *currDiag;
static Tgen *tgDiag;
static TRSF TCprev;
static TRSF TLprev;
static JntVec jcPrev;
static JntVec jdrivePrev;
static DSPL cdrivePrev;
static int pathTypePrev;
static int segCntPrev;
static int cntPrev;
static unsigned long confPrev;

checkMoveInit (tg)
Tgen *tg;
{
	currDiag = tgenGetDiag(tg);
	tgDiag = tg;
	zeroDspl (&cdrivePrev);
	zeroVf (jdrivePrev.v, NJNTS);
	cntPrev = 0;
	segCntPrev = 1;

	if (currDiag)
	 { TCprev = currDiag->T6;
	   identTrsf (&TLprev);
	   jcPrev = currDiag->j6;
	   confPrev = currDiag->conf;
	   pathTypePrev = 0;	/* doesn't matter since we're stopped */
	 }
}

checkMoveC (loc, segCnt, transCnt, TB, TC, conf, v1p)
int loc, segCnt, transCnt;
TRSF *TB, *TC;
unsigned long conf;
DIFF *v1p;
{
	TRSF TL;

	identTrsf (&TL);
	checkMoveT (loc, segCnt, transCnt, TB, TC, &TL, conf, v1p);
}

checkMoveT (loc, segCnt, transCnt, TB, TC, TL, conf, v1p)
int loc, segCnt, transCnt;
TRSF *TB, *TC, *TL;
unsigned long conf;
DIFF *v1p;
{
	TgenDiag *diag;
	TRSF TD, TCinv;
	TRSF T1, T2, TX, T6;
	DSPL cdrive;
	DIFF v2;
	DIFF delVel;
	MotionControl *mctl;
	double motionTime;
	int cnt;

	if (currDiag == NULL)
	 { bail (loc, "no more motions");
	 }
	if ((mctl = currDiag->mctl) == NULL)
	 { bail (loc, "no motion control desc");
	 }
	if (mctl->segCnt != segCnt)
	 { fprintf (stderr, "diag segCnt=%d segCnt=%d\n", 
			     mctl->segCnt, segCnt);
	   bail (loc, "bad segCnt");
	 }
	if (mctl->transCnt != transCnt)
	 { fprintf (stderr, "diag transCnt=%d transCnt=%d\n", 
			     mctl->transCnt, transCnt);
	   bail (loc, "bad transCnt");
	 }

	motionTime = tgDiag->sampInterval*segCnt;

	multLiTrsf (&TD, TC, TB);
	trsfToDspl (&cdrive, &TD);
	if (normVect(&cdrive.p) < TGEN_MM_EPSILON)
	 { zeroVect (&v2.t);
	   zeroVect (&cdrive.p);
	 }
	else
	 { scaleVect (&v2.t, -1/motionTime, &cdrive.p);
	 }
	if (ABS(cdrive.a) < TGEN_ANG_EPSILON)
	 { zeroVect (&v2.r);
	   cdrive.a = 0;
	 }
	else
	 { scaleVect (&v2.r, -ABS(cdrive.a)/motionTime, &cdrive.u);
	 }
	invertTrsf (&TCinv, TC);
	zeroVect (&TCinv.p);
	transDiff (&v2, &v2, &TCinv);

	subtractVect (&delVel.t, &v2.t, &v1p->t);
	subtractVect (&delVel.r, &v2.r, &v1p->r);

	cnt = -transCnt;	
	for (diag=currDiag->next; diag; diag=diag->next)
	 { cnt++;
	   
	   computeLinearMotionCart (&T2, TC, &cdrive, cnt/(double)segCnt);
	   if (cnt < transCnt)
	    { double s, transTime;

	      cntPrev++;
	      if (pathTypePrev == TGEN_CART_PATH)
	       { computeLinearMotionCart (&T1, &TCprev, &cdrivePrev,
					       cntPrev/(double)segCntPrev);
		 multRiTrsf (&T1, &T1, &TLprev);
		 multTrsf (&T1, &T1, TL);
	       }
	      else if (normVf((float*)v1p, 6) == 0)
	       { T1 = *TB;
	       }
	      else
	       { JntVec j1;
		 computeLinearMotionJnt (&j1, &jcPrev, &jdrivePrev,
					      cntPrev/(double)segCntPrev);
		 fwdKinematics (&T1, (unsigned long*)0, j1.v, 
				      tgDiag->kyn, (void*)0);
		 multTrsf (&T1, &T1, TL);
	       }
	      s = (cnt+transCnt)/(double)(2*transCnt);
	      transTime = 2*transCnt*tgDiag->sampInterval;
	      if (dbgHook1)
	       { printTrsf ("T1> %10.5m\n", &T1);
		 printTrsf ("T2> %10.5m\n", &T2);
		 printTrsf ("TB> %10.5m\n", TB);
		 printVf ("VP> %10.5f\n", (float*)v1p, 6);
		 printVf ("VC> %10.5f\n", (float*)&v2, 6);
	       }
	      blendRot (&TX, &T1, &v1p->r, &T2, &v2.r, TB, s, transTime);
	      blendVf ((float*)&TX.p, (float*)&T1.p, (float*)&v1p->t, 
		       (float*)&T2.p, (float*)&v2.t, s, transTime, 3);
	    }
	   else
	    { TX = T2;
	    }
	   multRiTrsf (&T6, &TX, TL);
	   if (!sameTrsf (&T6, &diag->T6))
	    { fprintf (stderr, "cnt=%d\n", cnt);
	      fprintTrsf (stderr, "    T6=%m\n", &T6);
	      fprintTrsf (stderr, "diagT6=%m\n", &diag->T6);
	      bail (loc, "position mismatch");
	    }
	   if (diag->mctl)
	    { break;
	    }
	 }
	if (diag)
	 { if (cnt != segCnt-diag->mctl->transCnt)
	    { bail (loc, "wrong cnt at new motion");
	    }
	 }
	TCprev = *TC;
	TLprev = *TL;
	confPrev = conf;
	cdrivePrev = cdrive;
	pathTypePrev = TGEN_CART_PATH;
	segCntPrev = segCnt;
	cntPrev = cnt;

	currDiag = diag;
}

checkMoveJ (loc, segCnt, transCnt, jb, jc, v1p)
int loc, segCnt, transCnt;
JntVec *jb, *jc;
JntVec *v1p;
{
	TgenDiag *diag;
	JntVec jd;
	JntVec j1, j2, j6;
	JntVec jdrive;
	JntVec v2, delVel;
	MotionControl *mctl;
	double motionTime;
	int cnt, j;

	if (currDiag == NULL)
	 { bail (loc, "no more motions");
	 }
	if ((mctl = currDiag->mctl) == NULL)
	 { bail (loc, "no motion control desc");
	 }
	if (mctl->segCnt != segCnt)
	 { fprintf (stderr, "diag segCnt=%d segCnt=%d\n", 
			     mctl->segCnt, segCnt);
	   bail (loc, "bad segCnt");
	 }
	if (mctl->transCnt != transCnt)
	 { fprintf (stderr, "diag transCnt=%d transCnt=%d\n", 
			     mctl->transCnt, transCnt);
	   bail (loc, "bad transCnt");
	 }

	motionTime = tgDiag->sampInterval*segCnt;

	subtractVf (jdrive.v, jc->v, jb->v, NJNTS);
	for (j=0; j<NJNTS; j++)
	 { if (ABS(jdrive.v[j]) < TGEN_ANG_EPSILON)
	    { jdrive.v[j] = 0;
	      v2.v[j] = 0;
	    }
	   else
	    { v2.v[j] = jdrive.v[j]/motionTime;
	    }
	 }
	subtractVf (delVel.v, v2.v, v1p->v, NJNTS);

	cnt = -transCnt;	
	for (diag=currDiag->next; diag; diag=diag->next)
	 { cnt++;
	   
	   computeLinearMotionJnt (&j2, jc, &jdrive, cnt/(double)segCnt);

	   if (cnt < transCnt)
	    { double s, transTime;

	      cntPrev++;
	      if (pathTypePrev == TGEN_JNT_PATH)
	       { computeLinearMotionJnt (&j1, &jcPrev, &jdrivePrev,
					       cntPrev/(double)segCntPrev);
	       }
	      else if (normVf(v1p->v, NJNTS) == 0)
	       { j1 = *jb;
	       }
	      else
	       { TRSF T1;
		 computeLinearMotionCart (&T1, &TCprev, &cdrivePrev,
					      cntPrev/(double)segCntPrev);
		 multRiTrsf (&T1, &T1, &TLprev);
		 invKinematics (j1.v, &T1, confPrev, diag->prev->j6.v,
				      tgDiag->kyn, tgDiag->jls, (void*)0);
	       }
	      s = (cnt+transCnt)/(double)(2*transCnt);
	      transTime = 2*transCnt*tgDiag->sampInterval;
	      blendVf (j6.v, j1.v, v1p->v, j2.v, v2.v, s, transTime, NJNTS);

	      if (dbgHook1)
	       { JntVec delJvel;

		 subtractVf (delJvel.v, v2.v, v1p->v, NJNTS);
		 printJf ("j1>      %10.5f\n", j1.v, NJNTS);
		 printJf ("j2>      %10.5f\n", j2.v, NJNTS);
		 printJf ("delJvel> %10.5f\n", delJvel.v, NJNTS);
		 printJf ("jvel>    %10.5f\n", v2.v, NJNTS);
		 printJf ("prevJvel>%10.5f\n", v1p->v, NJNTS);
	       }
	    }
	   else
	    { j6 = j2;
	    }
	   if (!sameVf (j6.v, diag->j6.v, 3))
	    { JntVec j6Deg, diagJ6Deg;
	      scaleVf (j6Deg.v, RADTODEG, j6.v, NJNTS);
	      scaleVf (diagJ6Deg.v, RADTODEG, diag->j6.v, NJNTS);
	      fprintf (stderr, "cnt=%d\n", cnt);
	      fprintVf (stderr, "     %12.7f\n", j6Deg.v, 3);
	      fprintVf (stderr, "diag %12.7f\n", diagJ6Deg.v, 3);
	      bail (loc, "position mismatch");
	    }
	   if (diag->mctl)
	    { break;
	    }
	 }
	if (diag)
	 { if (cnt != segCnt-diag->mctl->transCnt)
	    { bail (loc, "wrong cnt at new motion");
	    }
	 }
	jcPrev = *jc;
	jdrivePrev = jdrive;
	pathTypePrev = TGEN_JNT_PATH;
	segCntPrev = segCnt;
	cntPrev = cnt;

	currDiag = diag;
}

checkMoveEnd (loc)
int loc;
{
	if (currDiag != NULL)
	 { bail (loc, "extra functions");
	 }
}

main()
{
	Tgen *tg;
	TRSF T;
	JntVec j6;
	int j;

	identTrsf (&T);
	zeroVf (j6.v, NJNTS);

	/* Initialize a Cartesian only motion */

	tg = tgenAlloc ((char*)0);
	tgenInitCart (tg, 0.02, (TRSF*)0, 0);

	errorTest (101, tgenSetPath (tg, 20), TGBadArg);
	clearTest (102, tgenSetPath (tg, TGEN_CART_PATH));

	 { double tvelSave, rvelSave, tvel, rvel;

	   tgenGetCartVel (tg, &tvelSave, &rvelSave);
	   errorTest (103, tgenSetCartVel (tg,   0.0, -1.0), TGBadArg);
	   errorTest (104, tgenSetCartVel (tg,   1.0,  0.0), TGBadArg);
	   clearTest (105, tgenSetCartVel (tg,  24.0, -1.0));
	   tgenGetCartVel (tg, &tvel, &rvel);
	   if (!sameFloat(tvel, 24.0) || !sameFloat(rvel, rvelSave))
	    { bail (106, "bad setCartVel");
	    }
	   clearTest (107, tgenSetCartVel (tg,  -1.0, 333.0));
	   tgenGetCartVel (tg, &tvel, &rvel);
	   if (!sameFloat(tvel, 24.0) || !sameFloat(rvel, 333.0))
	    { bail (108, "bad setCartVel");
	    }
	   clearTest (109, tgenSetCartVel (tg,  tvelSave, rvelSave));
	 }

	 { double taccSave, raccSave, tacc, racc;

	   tgenGetCartAcc (tg, &taccSave, &raccSave);
	   errorTest (123, tgenSetCartAcc (tg,   0.0, -1.0), TGBadArg);
	   errorTest (124, tgenSetCartAcc (tg,   1.0,  0.0), TGBadArg);
	   clearTest (125, tgenSetCartAcc (tg,  24.0, -1.0));
	   tgenGetCartAcc (tg, &tacc, &racc);
	   if (!sameFloat(tacc, 24.0) || !sameFloat(racc, raccSave))
	    { bail (126, "bad setCartAcc");
	    }
	   clearTest (127, tgenSetCartAcc (tg,  -1.0, 333.0));
	   tgenGetCartAcc (tg, &tacc, &racc);
	   if (!sameFloat(tacc, 24.0) || !sameFloat(racc, 333.0))
	    { bail (128, "bad setCartAcc");
	    }
	   clearTest (129, tgenSetCartAcc (tg,  taccSave, raccSave));
	 }

	tgenInitJnt (tg, 0.02, (JntVec*)0);
	   
	 { double jvelSave, jvel;
	   int j;

	   jvelSave = tgenGetJntVel (tg, 2);
	   errorTest (142, tgenSetJntVel (tg,   0.0,  2), TGBadArg);
	   errorTest (144, tgenSetJntVel (tg, 111.0, NJNTS), TGBadArg);
	   clearTest (145, tgenSetJntVel (tg, 111.0, 2));
	   errorTest (146, (int)tgenGetJntVel (tg, -1), TGBadArg);
	   errorTest (147, (int)tgenGetJntVel (tg, NJNTS), TGBadArg);
	   jvel = tgenGetJntVel (tg, 2);
	   if (!sameFloat (jvel, 111.0))
	    { bail (148, "bad setJntVel");
	    }
	   for (j=0; j<NJNTS; j++)
	    { if (j != 2 && sameFloat (tgenGetJntVel (tg, j), 111.0))
	       { bail (149, "bad setJntVel");
	       }
	    }
	   clearTest (150, tgenSetJntVel (tg, 111.0, -1));
	   for (j=0; j<NJNTS; j++)
	    { if (!sameFloat (tgenGetJntVel (tg, j), 111.0))
	       { bail (151, "bad setJntVel");
	       }
	    }
	 }

	 { double jaccSave, jacc;
	   int j;

	   jaccSave = tgenGetJntAcc (tg, 2);
	   errorTest (142, tgenSetJntAcc (tg,   0.0,  2), TGBadArg);
	   errorTest (144, tgenSetJntAcc (tg, 111.0, NJNTS), TGBadArg);
	   clearTest (145, tgenSetJntAcc (tg, 111.0, 2));
	   errorTest (146, (int)tgenGetJntAcc (tg, -1), TGBadArg);
	   errorTest (147, (int)tgenGetJntAcc (tg, NJNTS), TGBadArg);
	   jacc = tgenGetJntAcc (tg, 2);
	   if (!sameFloat (jacc, 111.0))
	    { bail (148, "bad setJntAcc");
	    }
	   for (j=0; j<NJNTS; j++)
	    { if (j != 2 && sameFloat (tgenGetJntAcc (tg, j), 111.0))
	       { bail (149, "bad setJntAcc");
	       }
	    }
	   clearTest (150, tgenSetJntAcc (tg, 111.0, -1));
	   for (j=0; j<NJNTS; j++)
	    { if (!sameFloat (tgenGetJntAcc (tg, j), 111.0))
	       { bail (151, "bad setJntAcc");
	       }
	    }
	 }

	tgenFree (tg);
	if (tgenAlloc ("falsurmajager"))
	 { bail (200, "tgenAlloc worked with a nonexistent robot\n");
	 }
	clearErrors();
	if ((tg = tgenAlloc ("puma560")) == NULL)
	 { bail (201, "tgenAlloc failed with a good robot\n");
	 }

	 { JntVec j6;
	   TRSF T6;
	   unsigned long conf;

	   copyVf (j6.v, tg->jls->bestAngles, NJNTS);
	   fwdKinematics (&T6, &conf, j6.v, tg->kyn, (void*)0);

	   clearTest (202, tgenInitJnt (tg, 0.02, &j6));
	   clearTest (203, tgenInitCart (tg, 0.02, &T6, conf));
	   j6.v[0] += 10000*DEGTORAD;
	   errorTest (204, tgenInitJnt (tg, 0.02, &j6), TGJntLimit);
	   j6.v[0] -= 10000*DEGTORAD;
	   T6.p.x += 10000;
	   errorTest (205, tgenInitCart (tg, 0.02, &T6, 0), TGWorkspaceLimit);
	   T6.p.x -= 10000;

	   tgenInitJnt (tg, 0.02, (JntVec*)0);
	   if (!sameVf (j6.v, tg->j6.v, NJNTS))
	    { bail (206, "bad default joints");
	    }
	   tgenInitCart (tg, 0.02, (TRSF*)0, 0);
	   if (!sameTrsf (&T6, &tg->T6) || conf != tg->conf)
	    { bail (207, "bad default T6");
	    }
	 }

	tgenFree (tg);
#if 1
	tg = tgenAlloc ((char*)0);
	tgenInitCart (tg, 0.02, (TRSF*)0, 0);

	tgenEnableDiag (tg, 1);
	
	movec (tg,   0, 200, 90,  0,  0);
       	movec (tg, 200, 200,  0, 90,  0);
       	movec (tg, 200,   0,  0,  0, 90);
	movec (tg,   0,   0,  0,  0,  0);
	tgenStop (tg, 1.0);

	tgenSetCartVel (tg, -1.0, 1800.0);
	tgenSetCartAcc (tg, -1.0, 40*360.0);

	movec (tg,   0, 200, 90,  0,  0);
       	movec (tg, 200, 200,  0, 90,  0);
       	movec (tg, 200,   0,  0,  0, 90);
	movec (tg,   0,   0,  0,  0,  0);
	tgenStop (tg, 1.0);

	tgenSetCartAcc (tg, 1000.0, -1.0);

	movec (tg,   0, 200, 90,  0,  0);
       	movec (tg, 200, 200,  0, 90,  0);
       	movec (tg, 200,   0,  0,  0, 90);
	movec (tg,   0,   0,  0,  0,  0);
	tgenStop (tg, 1.0);

	tgenSetCartVel (tg, 500, 360);
	tgenSetCartAcc (tg, 5000, 2*1440);

	movec (tg,   0, 200, 90,  0,  0);
	tgenStop (tg, 0);
	errorTest (209, tgenSetPath (tg, TGEN_JNT_PATH), TGNoJntMotion);
       	movec (tg, 200, 200,  0, 90,  0);
	tgenStop (tg, 0.02);
	clearTest (210, tgenSetPath (tg, TGEN_CART_PATH));
       	movec (tg, 200,   0,  0,  0, 90);
	tgenStop (tg, 0.06);
	movec (tg,   0,   0,  0,  0,  0);
	tgenStop (tg, 0);

	tgenSetJntVel (tg, 100, -1);
	tgenSetJntAcc (tg, 500, -1);

	movej (tg,  0, 45, 0);
	movej (tg, 90, 90, 0);
	movej (tg, 20, 60, 0);
	movej (tg,  0,  0, 0);
	tgenStop (tg, 1);

	tgenSetJntVel (tg, 200, 1);

	movej (tg,  0, 45, 0);
	movej (tg, 90, 90, 0);
	movej (tg, 20, 60, 0);
	movej (tg,  0,  0, 0);
	tgenStop (tg, 1);

	tgenSetJntVel (tg, 100, 1);

	movej (tg,  0, 45, 0);
	tgenStop (tg, 0);
	movej (tg, 90, 90, 0);
	tgenStop (tg, 0.02);
	movej (tg, 20, 60, 0);
	tgenStop (tg, 0.22);
	movej (tg,  0,  0, 0);

	movec (tg,  0, 200, 90,  0,  0);
	movej (tg,  0, 45, 0);

	clearTest (211, tgenRun (tg, -1));
#endif
	 { TRSF TB0, TB1, TB2, TB3;
	   TRSF TB0inv, TB1inv, TB2inv, TB3inv;
	   DIFF  v0,  v1,  v2,  v3,  v4, v5;
	   TRSF TC1, TC2, TL1, TL2, TC0;
	   TRSF TD1, TD2;
	   double sinAng, cosAng;
	   int cnt1, cnt2, cnt3, cnt4;
	   double len;
	   int j;


	   JntVec jb0, jb1, jb2, jb3;
	   JntVec jv0, jv1, jv2, jv3, jv4;
	   JntVec jd0, jd1;

	   JntVec j60, j61, j62, j63;
	   unsigned long conf;

#if 1	   
	   identTrsf (&TB0);
	   zeroDiff (&v0);
	   zeroVf (j60.v, NJNTS);

	   xyzToTrsf (&TB1, 0, 200, 0);
	   rpyToTrsf (&TB1, 90, 0, 0);
	   setVect (&v1.t, 0, 1, 0, 400);
	   setVect (&v1.r, 0, 0, 1, 180*DEGTORAD);

	   xyzToTrsf (&TB2, 200, 200, 0);
	   rpyToTrsf (&TB2, 0, 90, 0);
	   setVect (&v2.t, 1, 0, 0, 294.1176);
	   setVect (&v2.r, -1, 1, -1, 120./.68*DEGTORAD/sqrt(3));

	   xyzToTrsf (&TB3, 200, 0, 0);
	   rpyToTrsf (&TB3, 0, 0, 90);
	   setVect (&v3.t, 0, -1, 0, 294.1176);
	   setVect (&v3.r, 1, -1, -1, 120./.68*DEGTORAD/sqrt(3));

	   setVect (&v4.t, -1, 0, 0, 400);
	   setVect (&v4.r, -1, 0, 0, 180*DEGTORAD);

	   checkMoveInit (tg);

	   /* rotation velocity dominates */
	   checkMoveC (300, 25, 4, &TB0, &TB1, 0, &v0);
	   checkMoveC (301, 34, 6, &TB1, &TB2, 0, &v1);
	   checkMoveC (302, 34, 6, &TB2, &TB3, 0, &v2);
	   checkMoveC (303, 25, 6, &TB3, &TB0, 0, &v3);

	   checkMoveC (304, 50, 4, &TB0, &TB0, 0, &v4);	/* stop */

	   setVect (&v1.t, 0, 1, 0, 500);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD/.4);

	   setVect (&v2.t, 1, 0, 0, 500);
	   setVect (&v2.r, -1, 1, -1, 120./.4*DEGTORAD/sqrt(3));

	   setVect (&v3.t, 0, -1, 0, 500);
	   setVect (&v3.r, 1, -1, -1, 120./.4*DEGTORAD/sqrt(3));

	   setVect (&v4.t, -1, 0, 0, 500);
	   setVect (&v4.r, -1, 0, 0, 90*DEGTORAD/.4);

	   checkMoveC (305, 20, 3, &TB0, &TB1, 0, &v0);
	   checkMoveC (306, 20, 4, &TB1, &TB2, 0, &v1);
	   checkMoveC (307, 20, 4, &TB2, &TB3, 0, &v2);
	   checkMoveC (308, 20, 4, &TB3, &TB0, 0, &v3);

	   checkMoveC (309, 50, 3, &TB0, &TB0, 0, &v4);	/* stop */

	   setVect (&v1.t, 0, 1, 0, 312.5);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD/.64);

	   setVect (&v2.t, 1, 0, 0, 312.5);
	   setVect (&v2.r, -1, 1, -1, 120./.64*DEGTORAD/sqrt(3));

	   setVect (&v3.t, 0, -1, 0, 312.5);
	   setVect (&v3.r, 1, -1, -1, 120./.64*DEGTORAD/sqrt(3));

	   setVect (&v4.t, -1, 0, 0, 312.5);
	   setVect (&v4.r, -1, 0, 0, 90*DEGTORAD/.64);

	   checkMoveC (310, 32, 8, &TB0, &TB1, 0, &v0);
	   checkMoveC (311, 32, 12, &TB1, &TB2, 0, &v1);
	   checkMoveC (312, 32, 12, &TB2, &TB3, 0, &v2);
	   checkMoveC (313, 32, 12, &TB3, &TB0, 0, &v3);

	   checkMoveC (314, 50, 8, &TB0, &TB0, 0, &v4);	/* stop */

	   setVect (&v1.t, 0, 1, 0, 500);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD/.4);

	   setVect (&v2.t, 1, 0, 0, 500);
	   setVect (&v2.r, -1, 1, -1, 120./.4*DEGTORAD/sqrt(3));

	   setVect (&v3.t, 0, -1, 0, 500);
	   setVect (&v3.r, 1, -1, -1, 120./.4*DEGTORAD/sqrt(3));

	   setVect (&v4.t, -1, 0, 0, 500);
	   setVect (&v4.r, -1, 0, 0, 90*DEGTORAD/.4);

	   checkMoveC (315, 20, 3, &TB0, &TB1, 0, &v0);
	   checkMoveC (316,  6, 3, &TB1, &TB1, 0, &v1);	/* stop 0 */
	   checkMoveC (317, 20, 3, &TB1, &TB2, 0, &v0);
	   checkMoveC (318,  6, 3, &TB2, &TB2, 0, &v2); /* stop 1 */
	   checkMoveC (319, 20, 3, &TB2, &TB3, 0, &v0);
	   checkMoveC (320,  6, 3, &TB3, &TB3, 0, &v3); /* stop 2 */
	   checkMoveC (321, 20, 3, &TB3, &TB0, 0, &v0);

	   checkMoveC (322,  8, 3, &TB0, &TB0, 0, &v4);

	   setJntVec (&jb0,  0.0,  0.0, 0.0, 0,0,0, 1);
	   setJntVec (&jb1,  0.0, 45.0, 0.0, 0,0,0, 1);
	   setJntVec (&jb2, 90, 90, 0, 0,0,0, 1);
	   setJntVec (&jb3, 20, 60, 0, 0,0,0, 1);

	   zeroVf (jv0.v, NJNTS);
	   setJntVec (&jv1,   0, 45, 0, 0,0,0, 1/(.02*23));
	   setJntVec (&jv2,  90, 45, 0, 0,0,0, 1/(.02*45));
	   setJntVec (&jv3, -70,-30, 0, 0,0,0, 1/(.02*35));
	   setJntVec (&jv4, -20,-60, 0, 0,0,0, 1/(.02*30));

	   checkMoveJ (400, 23, 5, &jb0, &jb1, &jv0);
	   checkMoveJ (401, 45, 5, &jb1, &jb2, &jv1);
	   checkMoveJ (402, 35,10, &jb2, &jb3, &jv2);
	   checkMoveJ (403, 30, 4, &jb3, &jb0, &jv3);
	   checkMoveJ (404, 50, 5, &jb0, &jb0, &jv4);	/* stop */

	   setJntVec (&jv1,   0, 45, 0, 0,0,0, 1/(.02*22));
	   setJntVec (&jv2,  90, 45, 0, 0,0,0, 1/(.02*45));
	   setJntVec (&jv3, -70,-30, 0, 0,0,0, 1/(.02*35));
	   setJntVec (&jv4, -20,-60, 0, 0,0,0, 1/(.02*25));

	   checkMoveJ (405, 22, 6, &jb0, &jb1, &jv0);
	   checkMoveJ (406, 45, 5, &jb1, &jb2, &jv1);
	   checkMoveJ (407, 35,10, &jb2, &jb3, &jv2);
	   checkMoveJ (408, 25, 4, &jb3, &jb0, &jv3);
	   checkMoveJ (409, 50, 6, &jb0, &jb0, &jv4);	/* stop */

	   setJntVec (&jv1,   0, 45, 0, 0,0,0, 1/(.02*23));
	   setJntVec (&jv2,  90, 45, 0, 0,0,0, 1/(.02*45));
	   setJntVec (&jv3, -70,-30, 0, 0,0,0, 1/(.02*35));
	   setJntVec (&jv4, -20,-60, 0, 0,0,0, 1/(.02*30));

	   checkMoveJ (410, 23, 5, &jb0, &jb1, &jv0);
	   checkMoveJ (411, 10, 5, &jb1, &jb1, &jv1);	/* stop */
	   checkMoveJ (412, 45, 5, &jb1, &jb2, &jv0);
	   checkMoveJ (413, 10, 5, &jb2, &jb2, &jv2);	/* stop */
	   checkMoveJ (414, 35, 5, &jb2, &jb3, &jv0);
	   checkMoveJ (415, 11, 5, &jb3, &jb3, &jv3);	/* stop */
	   checkMoveJ (416, 30, 5, &jb3, &jb0, &jv0);

	   checkMoveJ (417,  8, 5, &jb0, &jb0, &jv4);	/* auto stop */
	   checkMoveC (418, 20, 3, &TB0, &TB1, 0, &v0);
	   checkMoveC (419,  8, 3, &TB1, &TB1, 0, &v1); /* auto stop */
	   checkMoveJ (420, 23, 5, &jb0, &jb1, &jv0);

	   tgenInitCart (tg, 0.02, (TRSF*)0, 0);
	 
	   movec (tg, 0, 200, 90, 0, 0);
	   tgenRun (tg, 30);		/* run until idle */
	   movec (tg, 200, 200, 0, 90, 0);
	   tgenRun (tg, 19);		/* run until transitioning to idle */
	   movec (tg, 200, 0, 0, 0, 90);
	   tgenRun (tg, 10);		/* run until half done */
	   movec (tg, 0, 0, 0, 0, 0);
	   tgenRun (tg, 1);
	   movec (tg, 0, 200, 90, 0, 0);	   
	   movec (tg, 200, 200, 0, 90, 0);
	   tgenRun (tg, 1);
	   movec (tg, 200, 0, 0, 0, 90);
	   movec (tg, 0, 0, 0, 0, 0);
	   movec (tg, 0, 200, 90, 0, 0);

	   tgenSetMotionTime (tg, -1, 0);
	   movec (tg, 200, 200, 0, 90, 0);
	   movec (tg, 200, 0, 0, 0, 90);
	   tgenSetMotionTime (tg, 0, -1);
	   movec (tg, 0, 0, 0, 0, 0);
	   tgenSetMotionTime (tg, 1, -1);
	   movec (tg, 0, 200, 90, 0, 0);
	   movec (tg, 200, 200, 0, 90, 0);

	   clearTest (421, tgenRun (tg, -1));

	   checkMoveInit (tg);

	   setVect (&v1.t, 0, 1, 0, 500);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD/.4);
	   setVect (&v2.t, 1, 0, 0, 500);
	   setVect (&v2.r, -1, 1, -1, 120./.4*DEGTORAD/sqrt(3));
	   setVect (&v3.t, 0, -1, 0, 500);
	   setVect (&v3.r, 1, -1, -1, 120./.4*DEGTORAD/sqrt(3));
	   setVect (&v4.t, -1, 0, 0, 500);
	   setVect (&v4.r, -1, 0, 0, 90*DEGTORAD/.4);

	   checkMoveC (500, 20, 3, &TB0, &TB1, 0, &v0);
	   checkMoveC (501, 10, 3, &TB1, &TB1, 0, &v1);	/* idle stop */
	   checkMoveC (502, 20, 3, &TB1, &TB2, 0, &v0);
	   checkMoveC (503,  6, 3, &TB2, &TB2, 0, &v2); /* idle stop */
	   checkMoveC (504, 20, 3, &TB2, &TB3, 0, &v0);
	   checkMoveC (505,  6, 3, &TB3, &TB3, 0, &v3); /* idle stop */
	   checkMoveC (506, 20, 3, &TB3, &TB0, 0, &v0);
	   checkMoveC (507, 20, 4, &TB0, &TB1, 0, &v4);
	   checkMoveC (508, 20, 5, &TB1, &TB2, 0, &v1);
	   checkMoveC (509, 20, 5, &TB2, &TB3, 0, &v2);
	   checkMoveC (510, 20, 5, &TB3, &TB0, 0, &v3);
	   checkMoveC (511, 20, 4, &TB0, &TB1, 0, &v4);

	   /* motions with segment and/or transition times explicitly set */

	   setVect (&v1.t, 0, 1, 0, 500);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD/.4);
	   setVect (&v2.t, 1, 0, 0, 500);
	   setVect (&v2.r, -1, 1, -1, 120./.4*DEGTORAD/sqrt(3));
	   setVect (&v3.t, 0, -1, 0, 500);
	   setVect (&v3.r, 1, -1, -1, 120./.4*DEGTORAD/sqrt(3));
	   setVect (&v4.t, -1, 0, 0, 200/.3);
	   setVect (&v4.r, -1, 0, 0, 90*DEGTORAD/.3);

	   checkMoveC (512, 20, 0, &TB1, &TB2, 0, &v1);
	   checkMoveC (513, 20, 5, &TB2, &TB3, 0, &v2);
	   checkMoveC (514, 15, 5, &TB3, &TB0, 0, &v3);
	   checkMoveC (515, 50, 4, &TB0, &TB1, 0, &v4);

	   setVect (&v1.t, 0, 1, 0, 200);
	   setVect (&v1.r, 0, 0, 1, 90*DEGTORAD);

	   checkMoveC (516, 20, 4, &TB1, &TB2, 0, &v1);

	   tgenFree (tg);

	   tg = tgenAlloc ("puma560");
	   copyVf (j60.v, tg->jls->bestAngles, NJNTS);
	   tgenInitJnt (tg, 0.02, &j60);

	   fwdKinematics (&TB0, &conf, j60.v, tg->kyn, (void*)0);
	   TB1 = TB0;
	   TB2 = TB0;
	   TB3 = TB0;

	   TB1.p.y += 200.0;
	   TB2.p.x += 200.0;
	   TB2.p.y += 200.0;
	   TB3.p.x += 200.0;

	   invertTrsf (&TB1inv, &TB1);
	   zeroVect (&TB1inv.p);
	   invertTrsf (&TB2inv, &TB2);
	   zeroVect (&TB2inv.p);
	   invertTrsf (&TB3inv, &TB3);
	   zeroVect (&TB3inv.p);

	   invKinematics (j61.v, &TB1, conf, j60.v, tg->kyn, tg->jls,(void*)0);
	   invKinematics (j62.v, &TB2, conf, j60.v, tg->kyn, tg->jls,(void*)0);
	   invKinematics (j63.v, &TB3, conf, j60.v, tg->kyn, tg->jls,(void*)0);

	   tgenSetCartVel (tg, 200, 360);
	   tgenSetCartAcc (tg, 2000, 2*1440);
	   tgenSetJntVel (tg, 45.0, -1);
	   tgenSetJntAcc (tg, 300.0, -1);

	   tgenEnableDiag (tg, 1);

	   tgenMoveC (tg, &TB1, (TRSF*)0);
	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TB2, (TRSF*)0);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenMoveC (tg, &TB3, (TRSF*)0);
	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TB0, (TRSF*)0);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenStop (tg, 1.0);

	   if (tgenRun (tg, -1) < 0)
	    { printErrors();
	      bail (517, "tgenRun returns error\n");
	    }

	   setJntVec (&jv1, 0, -23.1804, 46.9506, 0, 
			    -23.7702, 0, 1/(.02*53));
	   fwdJacob (&v1, jv1.v, j61.v, tg->kyn);
	   transDiff (&v1, &v1, &TB1inv);

	   setVect (&v2.t, 200, 0, 0, 1/(.02*50));
	   setVect (&v2.r, 0, 0, 0, 0);
	   transDiff (&v2, &v2, &TB2);
	   invJacob (jv2.v, &v2, j62.v, tg->kyn, 1e-6);

	   setJntVec (&jv3, -4.0776, 30.7683, -61.9322, 
			    0, 31.1639, -4.0776, 1/(.02*69));
	   fwdJacob (&v3, jv3.v, j63.v, tg->kyn);
	   transDiff (&v3, &v3, &TB3inv);

	   setVect (&v4.t, -200, 0, 0, 1/(.02*50));
	   setVect (&v4.r, 0, 0, 0, 0);

	   checkMoveInit (tg);

	   checkMoveJ (520, 53, 4, &j60, &j61, &jv0);
	   checkMoveC (521, 50, 3, &TB1, &TB2, conf, &v1);
	   checkMoveJ (522, 69, 11, &j62, &j63, &jv2);
	   checkMoveC (523, 50, 4, &TB3, &TB0, conf, &v3);
	   checkMoveC (524, 50, 3, &TB0, &TB0, conf, &v4);

	   tgenClearDiag (tg);

	   conf = PUMA_DOWN_CONF|PUMA_FLIP_CONF;
	   tgenSetConf (tg, conf);
	   hook();
	   errorTest (525, tgenMoveC (tg, &TB0, (TRSF*)0), TGJntLimit);
	   conf = PUMA_FLIP_CONF;
	   tgenSetConf (tg, conf);

	   clearTest (526, tgenMoveC (tg, &TB0, (TRSF*)0));

	   tgenRun (tg, -1);

           checkMoveInit (tg);

	   invKinematics (j61.v, &TB0, conf, j60.v, tg->kyn, tg->jls,(void*)0);
/*
	   subtractVf (jv1.v, j61.v, j60.v, NJNTS);
	   scaleVf (jv1.v, jv1.v, NJNTS);
*/
	   checkMoveJ (527, 232, 4, &j60, &j61, &jv0);


	   tgenFree (tg);
#endif
	   /* Now do another test. This involves starting the
	      trajectory generator at the "best" angles, and then
	      doing the "handle" turning demo in several different
	      ways, to exercise the transition code between jnt
	      and Cartesian coordinates, with different tol frames.
	    */

#define ANGLE	20.0
#define LENGTH	100.0

	   tgenAlloc ("puma560");
	   copyVf (j60.v, tg->jls->bestAngles, NJNTS);
	   tgenInitJnt (tg, 0.02, &j60);	/* reset to best angles */
	   fwdKinematics (&TB0, &conf, j60.v, tg->kyn, (void*)0);

	   tgenSetCartVel (tg, 200, 180);
	   tgenSetCartAcc (tg, 2000, 2*1440);
	   tgenSetJntVel (tg, 45.0, -1);
	   tgenSetJntAcc (tg, 300.0, -1);

	   TB1 = TB0;
	   multTrsfRpy (&TB1, 0.0, ANGLE, 0.0);
	   TD1 = TB1;
	   TC1 = TB1;
	   multTrsfRpy (&TD1, 0.0, 0.0, -90.0);
	   multTrsfXyz (&TB1, 0.0, 0.0, -LENGTH);
	   TC0 = TB0;
	   multTrsfXyz (&TC0, 0.0, 0.0, LENGTH/2);
	   multTrsfXyz (&TC1, 0.0, 0.0, -LENGTH/2);
	   invKinematics (j61.v, &TB1, conf, j60.v, 
			  tg->kyn, tg->jls,(void*)0);

	   TB2 = TB0;
	   multTrsfRpy (&TB2, 0.0, -ANGLE, 0.0);
	   TD2 = TB2;
	   TC2 = TB2;
	   multTrsfRpy (&TD2, 0.0, 0.0, -90.0);
	   multTrsfXyz (&TB2, 0.0, 0.0, -LENGTH);
	   multTrsfXyz (&TC2, 0.0, 0.0, -LENGTH/2);
	   invKinematics (j62.v, &TB2, conf, j60.v, 
			  tg->kyn, tg->jls,(void*)0);

	   identTrsf (&TL1);
	   identTrsf (&TL2);
	   xyzToTrsf (&TL1, 0.0, 0.0, LENGTH/2);
	   xyzToTrsf (&TL2, 0.0, 0.0, LENGTH);
	   multTrsfRpy (&TL2, 0.0, 0.0, -90.0);

	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TC1, &TL1);
	   tgenStop (tg, 1.0);
	   tgenMoveC (tg, &TD2, &TL2);
	   tgenStop (tg, 1.0);
	   tgenMoveC (tg, &TB0, (TRSF*)0);
	   tgenStop (tg, 1.0);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenMoveC (tg, &TC1, &TL1);
	   tgenStop (tg, 1.0);
	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TD2, &TL2);
	   tgenStop (tg, 1.0);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenMoveC (tg, &TB0, (TRSF*)0);
	   tgenStop (tg, 1.0);
	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TC1, &TL1);
	   tgenMoveC (tg, &TD2, &TL2);
	   tgenMoveC (tg, &TB0, (TRSF*)0);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenMoveC (tg, &TC1, &TL1);
	   tgenSetPath (tg, TGEN_CART_PATH);
	   tgenMoveC (tg, &TD2, &TL2);
	   tgenSetPath (tg, TGEN_JNT_PATH);
	   tgenMoveC (tg, &TB0, (TRSF*)0);

	   tgenEnableDiag (tg, 1);

	   if (tgenRun (tg, -1) < 0)
	    { printErrors();
	      bail (547, "tgenRun");
	    }
	   checkMoveInit (tg);

	   sinAng = sin(DEGTORAD*ANGLE);
	   cosAng = cos(DEGTORAD*ANGLE);
	   len = (LENGTH/2)*sqrt((1+cosAng)*(1+cosAng) + sinAng*sinAng);

	   cnt1 = ceil(len/(.02*200));
/*	   cnt2 = ceil(2*ANGLE/(.02*180)); */
	   cnt2 = 12; /* because ANGLE is so small */

	   zeroDiff (&v0);
	   zeroVf (jv0.v, NJNTS);

	   setVect (&v1.t, -sinAng, 0, (1+cosAng), (LENGTH/2.0)/(.02*cnt1));
	   setVect (&v1.r, 0, -ANGLE, 0, DEGTORAD/(.02*cnt1));
	   setVect (&v2.t, cosAng, 0, -sinAng, 
			   2*ANGLE*DEGTORAD*LENGTH/(.02*cnt2));
	   setVect (&v2.r, 0, 2*ANGLE, 0, DEGTORAD/(.02*cnt2));

	   setVect (&v3.t, 0, 0, 0, 0);
	   setVect (&v3.r, 0, 2*ANGLE, 0, DEGTORAD/(.02*cnt2));

	   setVect (&v4.t, -sinAng, 0, -cosAng, LENGTH/(.02*cnt1));
	   setVect (&v4.r, 0, -ANGLE, 0, DEGTORAD/(.02*cnt1));

	   checkMoveT (550, cnt1, 3, &TC0, &TC1, &TL1, conf, &v0);
	   checkMoveT (551, 50, 3, &TC1, &TC1, &TL1, conf, &v1);
	   checkMoveT (552, cnt2, 2, &TD1, &TD2, &TL2, conf, &v0);
	   checkMoveT (553, 50, 2, &TD2, &TD2, &TL2, conf, &v3);
	   checkMoveC (554, cnt1, 3, &TB2, &TB0, conf, &v0);
	   checkMoveC (555, 50, 3, &TB0, &TB0, conf, &v4);

	   subtractVf (jd0.v, j61.v, j60.v, 6);
	   subtractVf (jd1.v, j60.v, j62.v, 6);

	   cnt1 = 27;
	   checkMoveJ (556, cnt1, 4, &j60, &j61, &jv0);
	   scaleVf (jv1.v, 1/(0.02*cnt1), jd0.v, 6);
	   checkMoveJ (557, 50, 4, &j61, &j61, &jv1);

	   cnt1 = 12;
	   checkMoveT (558, cnt1, 2, &TD1, &TD2, &TL2, conf, &v0);
	   checkMoveT (559, 50, 2, &TD2, &TD2, &TL2, conf, &v3);
	   cnt1 = 27;
	   checkMoveJ (560, cnt1, 4, &j62, &j60, &jv0);
	   scaleVf (jv2.v, 1/(0.02*cnt1), jd1.v, 6);
	   checkMoveJ (561, 50, 4, &j60, &j60, &jv2);

	   cnt1 = ceil(len/(.02*200));
	   checkMoveT (561, cnt1, 3, &TC0, &TC1, &TL1, conf, &v0);

	   setVect (&v1.t, -sinAng+cosAng*ANGLE*DEGTORAD, 
			   0, 
			   (1+cosAng) + sinAng*ANGLE*DEGTORAD,
			   (LENGTH/2.0)/(.02*cnt1));
	   checkMoveT (563, cnt2, 3, &TD1, &TD2, &TL2, conf, &v1);
	   checkMoveC (565, cnt1, 5, &TB2, &TB0, conf, &v2);
	   cnt1 = 27;
	   rotateDiff (&v4, &v4, &TB0);
	   invJacob (jv4.v, &v4, j60.v, tg->kyn, 1e-6);

	   checkMoveJ (566, cnt1, 4, &j60, &j61, &jv4);
	   cnt1 = 23;
	    { TRSF TD1inv;
	      fwdJacob (&v1, jv1.v, j61.v, tg->kyn);
	      transDiff (&v1, &v1, &TL2);
	      invertTrsf (&TD1inv, &TD1);
	      rotateDiff (&v1, &v1, &TD1inv);
	    }
	   checkMoveT (567, cnt1, 3, &TD1, &TD2, &TL2, conf, &v1);
	   setVect (&v2.t, cosAng, 0, -sinAng, 
			   2*ANGLE*DEGTORAD*LENGTH/(.02*cnt1));
	   setVect (&v2.r, 0, 2*ANGLE, 0, DEGTORAD/(.02*cnt1));

	   rotateDiff (&v2, &v2, &TB2);
	   invJacob (jv2.v, &v2, j62.v, tg->kyn, 1e-6);

	   cnt1 = 27;

	   checkMoveJ (568, cnt1, 20, &j62, &j60, &jv2);

/*	   checkMoveJ (556,  */
	 }

	printf ("\nRoutines OK\n\n");
}

hook()
{
}
