#ifdef Tsp_TARGET
/* Routines for working TGEN on a single transputer node */

#include <stdio.h>
#include <cdefs.h>
#include <conc.h>
#include <vector.h>
#include <varargInc.h>
#include "tgen.h"
#include "tgenTsp.h"
#include "servo.h"
#include "../h/stringDefs.h"

#define SEC_TO_TICKS	(1000000/64.)

#define REDUCED_FORCE	1

/* These are used to test the function timing */
int controlFxnCnt = 0;
int controlFxnTime = 0;
int robotComTime = 0;
int forceComTime = 0;
int useReducedForce = 0;

computeForce (force, ctrl)
FORCE *force;
TspTgenCtrl *ctrl;
{
	float ftemp[6];
	double accum;
	int temp[6];
/* Matrix is calibrated to output Newtons and Newton-mm, with an input
   in the range [-1,1] */

	static float matrix[6][6] = {
	   -3.3085,  54.5880,  -5.7139,  62.9138,   2.5368,-126.8194,
	    2.7955, -97.0946,   0.1177, 105.5729,  -0.3318,  -1.2085,
	  102.2091,  -1.9253, 108.9549,   4.8621, 102.3375,  -1.1971,

	   -1269.6,   -296.9,  -1515.8,    242.6,   2576.1,      1.1,
	    2207.4,   -157.5,  -2306.9,   -189.7,    -74.3,    374.9,
	    -141.7,   1359.3,   -173.3,   1624.3,    147.9,   1564.8 };

	/* This is a special matrix which eliminates the reading from
	   the second strain gauge and computes only fx,fy,fz and tx,tz
	   in a hand frame given by a z translation of 100 mm from the
	   sensor center. */
	static float rmatrix[6][6] = { 
	   21.3616,      0.0, -22.5817,  -0.0789,  -0.6506,   0.0870,
	  -41.0847,      0.0,  30.1203, 217.6167,   5.3376,-226.9344,
	  101.3390,      0.0, 109.5498,   7.0838, 102.4499,  -5.6730,

	   -1403.8,      0.0,  -1424.0,    585.3,   2593.4,   -689.2,
	    2136.2,      0.0,  -2258.2,     -7.9,    -65.1,      8.7,
	     472.6,      0.0,   -593.3,     55.7,     68.6,   4725.0 };

	int i, j;

	subtractVi (temp, ctrl->rawForceData, ctrl->rawForceBias, 6);

	if (useReducedForce)
	 { 
	   for (i=0; i<6; i++)
	    { accum = 0;
	      for (j=0; j<6; j++)
	       { accum += rmatrix[i][j]*temp[j]/512.0;
	       }
	      ftemp[i] = accum;
	    }
	 }
	else
	 { for (i=0; i<6; i++)
	    { accum = 0;
	      for (j=0; j<6; j++)
	       { accum += matrix[i][j]*temp[j]/512.0;
	       }
	      ftemp[i] = accum;
	    }
	 }

	/* Assume that process is running at high priority */
	copyVf ((float*)force, ftemp, 6);
	copyVi (ctrl->rawForceDel, temp, 6);
}

static void tspLock (tg)
Tgen *tg;
{
	tg->lockspace = ProcToHigh();
}

static void tspUnlock (tg)
Tgen *tg;
{
	if (tg->lockspace == 1)
	 { ProcToLow();
	 }
}

static void tspAbort (tg, code, fmt, ap)
Tgen *tg;
int code;
char *fmt;
va_list ap;
{
	printf ("TGEN real-time abort:\n");
	vprintf (fmt, ap);
	if (tg->ctrl)
	 { ProcStop (tg->ctrl->proc);
	 }
}

#define NAXES	8

static char buf[256];
static int servoStatus[NAXES] = 
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08 };

static unpack3byteInt (buf) 
unsigned char *buf;
{
	
	union _int {
	  unsigned char c[4];
	  int i;
	  } pack;
 
	pack.c[0] = buf[0];
	pack.c[1] = buf[1];
	pack.c[2] = buf[2];
	if (buf[2] & 0x80)
	 {  pack.c[3] = 0xff;
	 }
	else
	 {  pack.c[3] = 0;
	 }
	return (pack.i);
}
 
static int robotCom(tg, setpoint, position)
Tgen *tg;
int *setpoint, *position;
{
	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;
	int i, j;
	int t0;

	buf[0] = NORMAL_MODE;

	for (i=0, j=33; i < NAXES; i++, j +=2)
	 { buf[j] = (char) (servoStatus[i] & 0xFF);
	   buf[j+1] = (char) ((servoStatus[i] & 0xFF00) >> 8);
	 }
	BCOPY(setpoint, buf+1, 32);

	t0 = Time();
	ChanOut(ctrl->toServo, buf, MSG_LEN);
	ChanIn(ctrl->fromServo, buf, R_LEN_NORMAL_MODE);
	robotComTime += (Time()-t0);

	if (buf[0] != 0)
	 { tgenAbort (tg, TGComError, "robotCom(): com error\n");
	 }
	for (j=1, i=0; i<NAXES; j+=2, i++)
	 { servoStatus[i] = ((int) buf[j]) | (((int) buf[j+1]) << 8);
	 }
	BCOPY(buf+17, position, 32);
	return (1);
}

static int robotListen (tg, position)
Tgen *tg;
int *position;
{
	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;
	int i, j;

	buf[0] = LISTEN_MODE;

	for (i=0, j=33; i < NAXES; i++, j +=2)
	 { buf[j] = (char) (servoStatus[i] & 0xFF);
	   buf[j+1] = (char) ((servoStatus[i] & 0xFF00) >> 8);
	 }

	ChanOut(ctrl->toServo, buf, MSG_LEN);
	for (i = 0; i < R_LEN_LISTEN_MODE; i++)
	 { buf[i] = 0;
	 }

	ChanIn(ctrl->fromServo, buf, R_LEN_LISTEN_MODE);

	if (buf[0] != 0)
	 { tgenAbort (tg, TGComError, "robotListen(): com error\n");
	 }
	for (j=1, i=0; i<NAXES; j+=2, i++)
	 { servoStatus[i] = (((int) buf[j]) | (((int) buf[j+1]) << 8));
	 }
	BCOPY(buf+17, position, 32);
	return (1);
}

static int robotSetInterp (tg, i)
Tgen *tg;
int i; 
{
	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;

	buf[0] = INTERPOLATE;
	buf[1] = i;
	ChanOut (ctrl->toServo, buf, MSG_LEN);
	ChanIn (ctrl->fromServo, buf, R_LEN_ONE);
	if (buf[0] != 0)
	 { tgenAbort (tg, TGComError, "robotSetInterp(): com error\n");
	 }
	return (1);
}

static int robotSignal (tg)
Tgen *tg;
{
	/* Send an RCI request to start collecting information */

	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;

	ChanOutChar (ctrl->toServo, RCI_GATHER);
	ChanOutChar (ctrl->toServo, 0);
}

static int robotGather (tg, position)
Tgen *tg;
int *position;
{
	/* Gather an RCI information packet from the robot */

	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;
	unsigned char buf[256];
	unsigned char *bp;
	int len, j;

	len = ChanInChar (ctrl->fromServo);
	if (len != 0)
	 { ChanIn (ctrl->fromServo, buf, len);
	 }
	bp = buf;
	for (j=0; j<6; j++)
	 { position[j] = unpack3byteInt (bp);
	   bp += 3;
	 }
}

static robotInit (tg, interval, j6)
Tgen *tg;
double interval;
JntVec *j6;
{
	int encObs[NAXES];

	robotListen (tg, encObs);
	robotSetInterp (tg, (int)(interval*1000));
	encoderToAngle (j6->v, encObs, tg->jls);
}

static void controlFxn (proc, arg1)
Process *proc;
int arg1;
{
	Tgen *tg = (Tgen*)arg1;	/* XXX */
	TspTgenCtrl *ctrl = (TspTgenCtrl*)tg->ctrl;

	int ticks = 0;
	int resetTime = 0;
	int cnt = 0;

	SetTime (0);

	while (1)
	 { ticks += tg->sampInterval*1000000;
	   if (ticks > 2000000000)
	    { resetTime = 1;
	    }
	   ProcAfter (ticks); 
	   if (resetTime)
	    { SetTime (0);
	      ticks = 0;
	      resetTime = 0;
	    }
	   if (ctrl->exitReq)
	    { ctrl->exitReq = 0;
	      ProcStop (ctrl->proc);
	    }
	   else
	    { int encReq[NAXES];
	      int encObs[NAXES];
	      float angObs[6];
	      float rawFloatData[6];
	      int t0;

	      cnt++;
	      controlFxnCnt++;
	      t0 = Time();

	      robotSignal (tg);	      
#if 1
	      if (forceGet (rawFloatData) < 0)
	       { tgenAbort (TGComError, "forceGet() failed\n");
		 return;
	       }
#endif
	      forceComTime += (Time()-t0);

	      robotGather (tg, encObs);
	      encoderToAngle (tg->j6real.v, encObs, tg->jls);
	      fwdKinematics (&tg->T6real, &tg->confReal, tg->j6real.v,
			      tg->kyn, (void*)0);

	      scaleVf (rawFloatData, 512, rawFloatData, 6);	      
/* if ((cnt % 200) == 0) printVf ("Float: %4.0f\n", rawFloatData, 6); */
	      if (cnt == 1 || ctrl->resetBias)
	       { copyVfi (ctrl->rawForceBias, rawFloatData, 6);
	       }
	      copyVfi (ctrl->rawForceData, rawFloatData, 6);

	      computeForce (&tg->force, ctrl);
	      if (ctrl->resetBias)
	       { ctrl->resetBias = 0;
	       }
	      trajectoryCycle (tg);

	      /* now convert the joint angles into CRS plus coordinates */

	      angleToEncoder (encReq, tg->j6.v, tg->jls);
	      robotCom (tg, encReq, encObs);
#if 0
	      encoderToAngle (tg->j6real.v, encObs, tg->jls);
	      fwdKinematics (&tg->T6real, &tg->confReal, tg->j6real.v,
			      tg->kyn, (void*)0);
#endif
	      controlFxnTime += (Time()-t0);
	    }
	 }
}

tgenStart (tg, interval, toServo, fromServo, toKyn, fromKyn)
Tgen *tg;
double interval;
Channel *toServo;
Channel *fromServo;
Channel *toKyn;
Channel *fromKyn;
{
	JntVec j6;
	TspTgenCtrl *ctrl;

	/* create a real-time task to run the tgen */

	ctrl = ALLOC(TspTgenCtrl);
	ctrl->exitReq = 0;
	ctrl->sampCnt = interval*1000000;
	ctrl->toServo = toServo;
	ctrl->fromServo = fromServo;
	ctrl->toKyn = toKyn;
	ctrl->fromKyn = fromKyn;
	ctrl->proc = ProcAlloc (controlFxn, 0x2000, 1, (int)tg);
	tg->lock = tspLock;
	tg->unlock = tspUnlock;
	tg->abort = tspAbort;
	tg->ctrl = ctrl;

	forceSetChan (ctrl->toKyn, ctrl->fromKyn);
printf ("calling robotInit ...\n");
	robotInit (tg, interval, &j6);
printf ("done\n");
	if (tgenInitJnt (tg, interval, &j6) < 0)
	 { return (-1);
	 }
	tg->confReal = tg->conf;
	tg->T6real = tg->T6;
	tg->j6real = tg->j6;
printf ("here\n");
	ProcRunHigh (ctrl->proc);

	return (1);
}

tgenRelease (tg)
Tgen *tg;
{
	TspTgenCtrl *ctrl;

	ctrl = (TspTgenCtrl*)tg->ctrl;
	ctrl->exitReq = 1;
	while (ctrl->exitReq)
	   ;
	ProcFree (ctrl->proc);
	free (tg->ctrl);
	tg->lock = NULL;
	tg->unlock = NULL;
	tg->ctrl = NULL;

	return (1);
}

#endif /* Tsp_TARGET */


#if 0

/* Old matrix definitions supplied by Zebra. They are no longer
   valid, and are the transpose of the matrices which we use above. */

	static float offset[6] = {
	 -0.192770, -0.217239, -0.573698,  -0.363458,  -0.104008,  -0.289887
	};

	static float matrix[6][6] = {
	  0.071988,  0.063493,  2.114726,  43.420361, -23.565434,  -3.406176,
	 -2.209872,  1.220919, -0.051260,  10.088677,  19.407791,  30.649527,
	  0.004277,  0.124238,  2.098812,  -0.581925,  53.101963,  -1.113001,
	  0.065277, -2.413165,  0.059846, -20.353119,  -2.418571,  30.292025,
	 -0.062870,  0.099718,  1.979402, -45.001221, -23.365202,   0.183686,
	  2.220282,  1.196784,  0.017931,   9.056970, -17.949341,  32.436016 };

	static float offset[6] = {
	 -1.240662, -0.452871, -0.638534,  -0.407676,  -0.353692,  -0.344252
	};

	static float matrix[6][6] = {
	  0.078493, -0.252082, -2.173205, -48.675835,  26.142397,  -2.418778,
	  2.993643, -1.857505, -0.125321, -18.936853, -24.355850, -45.255215,
	  0.013218,  0.066148,  2.238842,   4.168168,  52.069267,  -0.277898,
	  0.068671,  3.816274, -0.092800,  30.195377,   3.565370, -42.148746,
	 -0.042646,  0.089120, -2.137797,  40.665928,  25.417624,   0.546231,
	  3.143781,  1.803730, -0.112082,  10.168342, -21.826309,  46.128845 };
#endif
