#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>

#include "rci.h"
#include "rciioctl.h"
#include "genrcireg.h"

#define STATIC

int genrcidevflag = 0;
int genrcicnt = 0;
int genrciwcnt = 0;

extern int genrci_internal_major;
extern GenrciSoftc genrcii[];

static void onTimeout();
static resetTimerQueue();
static queueTimerEvent();
static getTimerEvent();

static void onTimeout (caddr_t arg)
{
	/* Timeout routine. Implements a private 'hardclock' routine
	   that keeps track of the time ('tick') and queues events
	   (alarms and beat pulses) when they come due. */

	register GenrciSoftc *info = (GenrciSoftc*)arg;
	int req = 0;

	genrcicnt++;	/* DBG */

	info->tid = itimeout (onTimeout, arg, 1, plhi); /* reset timeout right away */
	info->tidset = 1; /* isn't this unnecessary? */
	info->tick++;

	if (info->alarmTick >= 0 && info->alarmTick <= info->tick)
	 { req |= RCI_ALARM_READY;
	   info->alarmTick = -1;
	 }
	if (info->beatTick >= 0 && info->beatTick <= info->tick)
	 { req |= RCI_BEAT_READY;
	   info->beatTick += info->beatPeriod;
	 }
	if (req)
	 { if (queueTimerEvent(&info->tq, req, info->tick))
	    { pollwakeup (info->pollhead, POLLRDNORM);
/*	      cmn_err (CE_CONT, "T: %x\n",info->pollhead); */
	      genrciwcnt++; /*DBG*/
	    }
	 }
}

STATIC int genrciinit ()
{
}

STATIC int genrciopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device can only be openned by one process at a time;
	   this mainly makes the software a bit easier to implement. */
	
      	GenrciSoftc *info;
	int unit;
	int s;

/*	info = &genrcii[0]; */
	info = &genrcii[geteminor(*dev)];

	s = splhi();
	if (info->flags & GENRCI_OPEN)
	 { splx(s);
	   return (EBUSY);
	 }
	else
	 { info->flags |= GENRCI_OPEN;
	 }
	splx(s);

	info->beatTick = -1;
	info->beatPeriod = 0;
	info->alarmTick = -1;
	info->tick = 0;
	info->tidset = 0;
	info->pollhead = phalloc (KM_SLEEP);
	resetTimerQueue(&info->tq,TQ_DEFAULT_SIZE);

	return (0);
}

STATIC int genrciclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	/* When closing the device, be sure to clear any pending 
	   timeouts. */

	GenrciSoftc *info;
	int s;

	info = &genrcii[geteminor(dev)];
	if (info->tidset)
	 { untimeout (info->tid);
	   info->tidset = 0;
	 }
	s = splhi();
	if (info->pollhead)
	 { phfree (info->pollhead);
	 }
	else
	 { cmn_err (CE_WARN, "xydclose: pollhead is NULL\n");
	 }
	info->flags &= ~(GENRCI_OPEN);
	splx(s);
	return (0);
}

STATIC int genrciread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* If there are any timer events pending, then return
	   the first one in the queue. Otherwise, return nothing. */

	GenrciSoftc *info;
	RciTimerEvent tevent;

	info = &genrcii[geteminor(dev)];
	if (getTimerEvent(&info->tq, &tevent))
	 { if (uiomove ((caddr_t)&tevent, sizeof(RciTimerEvent), 
					  UIO_READ, uiop))
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int genrciwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int genrcichpoll (dev_t dev, short events, int anyyet,
			 short *reventsp, struct pollhead **phpp)
{
	/* If there are no timer events pending, the queue the process
	   for a possible select wakeup. */

	GenrciSoftc *info;
	int ready, s;

	info = &genrcii[geteminor(dev)];

	/* What about races? */
	if ((events & (POLLIN|POLLRDNORM)) && !TIMER_QUEUE_EMPTY(&info->tq))
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
/*	   if (!anyyet) */
	    { *phpp = info->pollhead;
	    }
	 }
	return (0);
}

STATIC int genrciioctl (dev_t dev, int cmd, void *arg, int mode,
			cred_t *crp, int *rvalp)
{
	GenrciSoftc *info;

	info = &genrcii[geteminor(dev)];
	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_TIMER_ENABLE:
	    { int enable = (int)arg;

	      if (enable && !info->tidset)
	       { info->tid = timeout (onTimeout, (caddr_t)info, 1, plhi);
		 info->tidset = 1;
	       }
	      else if (!enable && info->tidset)
	       { untimeout (info->tid);
		 info->tidset = 0;
	       }
	      break;
	    }
	   case RCIIOC_TIMER_GET:
	    { int tickval;

	      tickval = info->tick; /* hope this is atomic */
	      if (copyout((char*)&tickval, (char*)arg, sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case RCIIOC_TIMER_SET:
	    { int tick = (int)arg;

	      if (tick < 0)
	       { return (EINVAL);
		 break;
	       }
	      info->tick = tick; /* hope this is atomic */
	      break;
	    }
	   case RCIIOC_BEAT_REQ:
	    { int period = (int)arg;
	      int s;

	      if (period < 0)
	       { return (EINVAL);
		 break;
	       }
	      s = splhi();
	      if (period)
	       { info->beatPeriod = period;
		 info->beatTick = info->tick + period;
	       }
	      else
	       { info->beatTick = -1;
		 info->beatPeriod = 0;
	       }
	      splx(s);
	      break;
	    }
	   case RCIIOC_ALARM_REQ:
	    { int ticks = (int)arg;
	      int s;

	      if (ticks < 0)
	       { return (EINVAL);
		 break;
	       }
	      s = splhi();
	      if (ticks <= info->tick)
	       { if (queueTimerEvent(&info->tq, RCI_ALARM_READY, info->tick))
		  { pollwakeup (info->pollhead, POLLRDNORM);
		  }
	       }
	      else
	       { info->alarmTick = ticks;
	       }
	      splx(s);

	      break;
	    }
	   case RCIIOC_SET_TIMERQ:
	    { int size = (int)arg;

	      if ((size < 1) || (size > TQ_MAX_SIZE))
	       { return (EINVAL);
		 break;
	       }
	      resetTimerQueue (&info->tq, size);
	      break;
	    }
	   case RCIIOC_GET_TIMERQ:
	    { int size;

	      size = info->tq.size;
	      if (copyout ((char*)&size, (char*)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
	 }
	return (0);
}

static resetTimerQueue (RciTimerQueue *tq, int size)
{
	/* Clear a timer event queue and set its size to the
	   specified value. */

	int s;

	s = splhi();
	tq->size = size;
	tq->head = 0;
	tq->tail = 0;
	splx(s);
}

static queueTimerEvent (RciTimerQueue *tq, int req, int tick)
{
	/* Queue a timer event, if there is room on the queue, by
	   advancing the tail and adding it there. */

	int newtail;

	newtail = tq->tail + 1;
	if (newtail > tq->size) newtail = 0;
	if (newtail != tq->head)
	 { tq->buf[tq->tail].req = req;
	   tq->buf[tq->tail].tick = tick;
	   tq->tail = newtail;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static getTimerEvent (RciTimerQueue *tq, RciTimerEvent *event)
{
	/* Remove a timer event from the queue, 
	   if there are any there. */

	if (tq->head != tq->tail)
	 { int newhead;
	   event->req = tq->buf[tq->head].req;
	   event->tick = tq->buf[tq->head].tick;
	   newhead = tq->head + 1;
	   if (newhead > tq->size) newhead = 0;
	   tq->head = newhead;
	   return (1);
	 }
	else
	 { return (0);
	 }
}
