#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#include "rcisys.h"
#include "rciioctl.h"
#include "genrcireg.h"

#define STATIC

static RciTimeoutFxns timeoutFxns;
static int use_sysclock;

static int reqTimeout (void(*fxn)(caddr_t), void *arg, int ticks)
{
	if (use_sysclock || timeoutFxns.req == 0)
	 { return (realtime_timeout (fxn, arg, ticks));
	 }
	else
	 { return ((*timeoutFxns.req)(timeoutFxns.arg, fxn, arg, ticks));
	 }
}

static void delTimeout (int id)
{
	if (use_sysclock || timeoutFxns.del == 0)
	 { untimeout (id);
	 }
	else
	 { (*timeoutFxns.del)(timeoutFxns.arg, id);
	 }
}

static void setTimeoutFxns (RciTimeoutFxns *fxns)
{
	if (fxns == 0)
	 { use_sysclock = 1;
	   timeoutFxns.req = 0;
	   timeoutFxns.del = 0;
	 }
	else
	 { timeoutFxns = *fxns;
	 }
}

int genrcidevflag = 0;
void *genrcii;

static void onTimeout();
static resetTimerQueue();
static queueTimerEvent();
static getTimerEvent();

/* Stuff relating to dynamic loading and autoconfiguration */

static int genrciidentify();
static int genrciattach();
static int genrcidetach();
static int genrcigetinfo();
static int genrciprobe();

STATIC int genrciopen(dev_t *devp, int flag, int otyp, cred_t *credp);
STATIC int genrciclose(dev_t dev, int flag, int otyp, cred_t *credp);
STATIC int genrciread(dev_t dev, struct uio *uiop, cred_t *credp);
STATIC int genrciwrite(dev_t dev, struct uio *uiop, cred_t *credp);
STATIC int genrciioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
STATIC int genrcichpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static struct cb_ops genrci_cb_ops = {
	genrciopen,
	genrciclose,
	nodev,
	nodev,
	nodev,
	genrciread,
	genrciwrite,
	genrciioctl,
	nodev,
	nodev,
	nodev,
	genrcichpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops genrci_ops = {
	DEVO_REV,
	0,
	genrcigetinfo,
	genrciidentify,
	genrciprobe,
	genrciattach,
	genrcidetach,
	nodev,
	&genrci_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"genrci driver",
	&genrci_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&genrcii, sizeof(GenrciSoftc), 4);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&genrcii);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&genrcii);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int genrciidentify (dip)
dev_info_t *dip;
{
	if (strcmp (ddi_get_name(dip), "genrci") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int genrcigetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	GenrciSoftc *info;
	dev_t dev;
	int minor, status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { dev = (dev_t) arg;
	      *result = (void*)getminor(dev);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(genrcii, getminor((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

genrciattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	GenrciSoftc *info;
	int minor;

	static char *minorNames[] = { "0", "1", "2", "3" };

	switch (cmd)
	 { case DDI_ATTACH:
	    { minor = ddi_get_instance(dip);
	      if (minor >= 4)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(genrcii, minor) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(genrcii, minor);
	      mutex_init (&info->mu, "genrci mutex", MUTEX_DRIVER,
			  (void*)info->iblock_cookie);
	      info->savedCallGate.s_base = 0;
	      info->savedCallGate.s_limacc = 0;
/*	      cv_init (&info->cv, "genrci cv", CV_DRIVER, NULL); */
	      info->dip = dip;
	      if (ddi_create_minor_node (dip, minorNames[minor], S_IFCHR, 
					 minor, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { goto failed;
	       }
	      ddi_report_dev(dip);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
failed:
/*	cv_destroy (&info->cv); */
	mutex_destroy (&info->mu);
	ddi_soft_state_free (genrcii, minor);
	return (DDI_FAILURE);
}

genrcidetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	GenrciSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (genrcii, ddi_get_instance(dip));
	      ddi_remove_minor_node (dip, NULL);
	      mutex_destroy (&info->mu);
	      ddi_soft_state_free (genrcii, ddi_get_instance(dip));
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

genrciprobe (dev_info_t *dip)
{
	return (DDI_PROBE_DONTCARE);
}

static void onTimeout (caddr_t arg)
{
	/* Timeout routine. Implements a private 'hardclock' routine
	   that keeps track of the time ('tick') and queues events
	   (alarms and beat pulses) when they come due. */

	register GenrciSoftc *info = (GenrciSoftc*)arg;
	int req = 0;

	mutex_enter (&info->mu);
	info->tid = reqTimeout (onTimeout, arg, 1); /* reset timeout right away */
	info->tidset = 1; /* isn't this unnecessary? */
	info->tick++;

	if (info->alarmTick >= 0 && info->alarmTick <= info->tick)
	 { req |= RCI_ALARM_READY;
	   info->alarmTick = -1;
	 }
	if (info->beatTick >= 0 && info->beatTick <= info->tick)
	 { req |= RCI_BEAT_READY;
	   info->beatTick += info->beatPeriod;
	 }
	if (req && queueTimerEvent(&info->tq, req, info->tick))
	 { mutex_exit (&info->mu);
	   pollwakeup (&info->pollhead, POLLRDNORM);
	 }
	else
	 { mutex_exit (&info->mu);
	 }
}

STATIC int genrciinit ()
{
}

STATIC int genrciopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device can only be openned by one process at a time;
	   this mainly makes the software a bit easier to implement. */
	
      	GenrciSoftc *info;
	int unit;
	int s;

	info = ddi_get_soft_state(genrcii,getminor(*dev));

	mutex_enter (&info->mu);
	if (info->flags & GENRCI_OPEN)
	 { mutex_exit (&info->mu);
	   return (EBUSY);
	 }
	else
	 { info->flags |= GENRCI_OPEN;
	 }
	mutex_exit (&info->mu);

	/* info is unprotected by locks here because the device 
	   is exclusive open and there won't be any timeout callbacks
	   coming in at this point */
	info->beatTick = -1;
	info->beatPeriod = 0;
	info->alarmTick = -1;
	info->tick = 0;
	info->tidset = 0;
	bzero ((char*)&info->pollhead, sizeof(struct pollhead));
	mutex_enter (&info->mu);
	resetTimerQueue(&info->tq,TQ_DEFAULT_SIZE);
	mutex_exit (&info->mu);

/*	foo(); */

	return (0);
}

STATIC int genrciclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	/* When closing the device, be sure to clear any pending 
	   timeouts. */

	GenrciSoftc *info;
	int s;

	info = ddi_get_soft_state(genrcii,getminor(dev));

	/* info->tidset can't be set false by timeout callback.
	   Therefore, since device is exclusive open, the following
	   code fragment is safe */
	if (info->tidset)
	 { delTimeout (info->tid);
	   info->tidset = 0;
	 }
	mutex_enter (&info->mu);
	info->flags &= ~(GENRCI_OPEN);
	mutex_exit (&info->mu);
	return (0);
}

STATIC int genrciread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* If there are any timer events pending, then return
	   the first one in the queue. Otherwise, return nothing. */

	GenrciSoftc *info;
	RciTimerEvent tevent;

	info = ddi_get_soft_state(genrcii,getminor(dev));
	if (getTimerEvent(&info->tq, &tevent))
	 { if (uiomove ((caddr_t)&tevent, sizeof(RciTimerEvent), 
					  UIO_READ, uiop))
	   
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int genrciwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int genrcichpoll (dev_t dev, short events, int anyyet,
			 short *reventsp, struct pollhead **phpp)
{
	/* If there are no timer events pending, the queue the process
	   for a possible select wakeup. */

	GenrciSoftc *info;
	int ready, s;

	info = ddi_get_soft_state(genrcii,getminor(dev));

	/* What about races? */
	if ((events & (POLLIN|POLLRDNORM)) && !TIMER_QUEUE_EMPTY(&info->tq))
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = &info->pollhead;
	    }
	 }
	return (0);
}

/* -------------------------------------------------------------- 
   The next few routines support the RCIIOC_ENABLE_IOPORTS icotl 
   which gives a user-level process access to the I/O ports. This
   is done by setting the IOPL level in the EFLAGS register to
   0x3. The only problem is that normally EFLAGS is restored off
   the stack when kernel routines return. That means that EFLAGS
   must be set *before* it is saved by the kernel. The way to do
   this is to temporarily install a patch right at the system
   call entry point -- which can be easily done by temporarily
   changing the call gate entry point for system calls to the
   patch location.

	RCIIOC_ENABLE_IOPORTS should be called twice: 
	first with an argument value of 1, to setup the patch,
	and second with an argument of 0, to execute and remove the patch
*/

#include <sys/cpuvar.h>	/* used by the RCIIOC_ENABLE_IOPORTS call */

static unsigned long getSegOffset (struct seg_desc *desc)
{
	/* Get the offset stored in a segment descriptor */

	return (((desc->s_limacc & 0xff000000)) +
		((desc->s_limacc & 0xff) << 16) + 
		((desc->s_base & 0xffff0000) >> 16));
}

static unsigned long getGateEntry (struct seg_desc *desc)
{
	/* Returns the entry point stored in a call gate.
	   Not used right now */

	return (((desc->s_limacc & 0xffff0000)) +
		((desc->s_base & 0xffff)));
}

void genrci_io_patch()
{
	/* Temporary kernel entry point which sets IOPL to 3
	   in the EFLAGS register, before jumping to the normal
	   kernel entry poiint (sys_call()) */

	extern sys_call();

	/* NOTE: EFLAGS register can only be read/set by pushing/popping
	   it on/off tyhe stack */

	asm ("genrci_io_patch_entry:");		/* entry point */
	asm ("pushf");				/* push low word of EFLAGS */
	asm ("orw	$0x3000,(%esp)");	/* set IOPL in pushed value */
	asm ("popf");				/* pop low word of EFLAGS */
	asm ("jmp	sys_call");		/* jump to normal entry point */
}

static void setGateEntry (struct seg_desc *new, struct seg_desc *old, 
		 unsigned long entry)
{
	/* Sets the entry point in a call gate */

	new->s_base = (old->s_base & 0xffff0000) | (entry & 0xffff);
	new->s_limacc = (old->s_limacc & 0xffff) | (entry & 0xffff0000);
}

static unsigned short getLDTReg()
{
	/* Gets the current local descriptor table register.
	   Not used right now */

	asm ("sldt %ax");
}

static int getEFLAGS()
{
	/* Get the low 16 bits of the EFLAGS register. Not used right now */

	asm ("pushf");
	asm ("popw %eax");
}

/* ------------------------------------------------------------ */

STATIC int genrciioctl (dev_t dev, int cmd, int arg, int mode,
			cred_t *crp, int *rvalp)
{
	GenrciSoftc *info;

	info = ddi_get_soft_state(genrcii,getminor(dev));
	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_TIMER_ENABLE:
	    { int enable = (int)arg;

	      if (enable && !info->tidset)
	       { /* no timeout callbacks can occur here */
		 info->tid = reqTimeout (onTimeout, (caddr_t)info, 1);
		 info->tidset = 1;
	       }
	      else if (!enable && info->tidset)
	       { /* code fragement is MT safe; see comment in genrciclose() */
		 delTimeout (info->tid);
		 info->tidset = 0;
	       }
	      break;
	    }
	   case RCIIOC_TIMER_GET:
	    { int tickval;

	      tickval = info->tick; /* hope this is atomic */
	      if (copyout((caddr_t)&tickval, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
	       }
	      break;
	    }
	   case RCIIOC_TIMER_SET:
	    { int tick = (int)arg;

	      if (tick < 0)
	       { return (EINVAL);
		 break;
	       }
	      mutex_enter (&info->mu);
	      info->tick = tick; /* hope this is atomic */
	      mutex_exit (&info->mu);
	      break;
	    }
	   case RCIIOC_BEAT_REQ:
	    { int period = (int)arg;
	      int s;

	      if (period < 0)
	       { return (EINVAL);
		 break;
	       }
	      mutex_enter (&info->mu);
	      if (period)
	       { info->beatPeriod = period;
		 info->beatTick = info->tick + period;
	       }
	      else
	       { info->beatTick = -1;
		 info->beatPeriod = 0;
	       }
	      mutex_exit (&info->mu);
	      break;
	    }
	   case RCIIOC_ALARM_REQ:
	    { int ticks = (int)arg;
	      int s;

	      if (ticks < 0)
	       { return (EINVAL);
		 break;
	       }
	      mutex_enter (&info->mu);
	      if (ticks <= info->tick)
	       { if (queueTimerEvent(&info->tq, RCI_ALARM_READY, info->tick))
		  { pollwakeup (&info->pollhead, POLLRDNORM);
		  }
	       }
	      else
	       { info->alarmTick = ticks;
	       }
	      mutex_exit (&info->mu);

	      break;
	    }
	   case RCIIOC_SET_TIMERQ:
	    { int size = (int)arg;

	      if ((size < 1) || (size > TQ_MAX_SIZE))
	       { return (EINVAL);
		 break;
	       }
	      mutex_enter (&info->mu);
	      resetTimerQueue (&info->tq, size);
	      mutex_exit (&info->mu);
	      break;
	    }
	   case RCIIOC_GET_TIMERQ:
	    { int size;

	      size = info->tq.size;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
	   case RCIIOC_GET_FXNS:
	    { RciFxnInterface fxni;

	      fxni.setTimeoutFxns = setTimeoutFxns;
	      if (copyout ((caddr_t)&fxni, (caddr_t)arg, sizeof(fxni)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
	   case RCIIOC_ENABLE_IOPORTS:
	    { struct seg_desc *callgate;
	      struct seg_desc *ldtdesc;
	      struct seg_desc newGate;
	      int iarg = (int)arg;

	      extern void *genrci_io_patch_entry();
	      extern cpu_t *cpu_list;

	      if (USER_SCALL & 0x4)
	       { /* call gate is in local descriptor table */
		 ldtdesc = &cpu_list->cpu_m.mcpu_gdt[LDTSEL>>3];
		 callgate = (struct seg_desc*)getSegOffset(ldtdesc);
	       }
	      else
	       { /* call gate is in global descriptor table */
		 callgate = &cpu_list->cpu_m.mcpu_gdt[USER_SCALL>>3];
	       }
	      if (iarg)
	       { info->savedCallGate = *callgate;
		 setGateEntry (&newGate, &info->savedCallGate,
			       (unsigned long)genrci_io_patch_entry);
		 callgate->s_base = newGate.s_base;
		 callgate->s_limacc = newGate.s_limacc;
	       }
	      else
	       { if (info->savedCallGate.s_base != 0)
		  { callgate->s_base = info->savedCallGate.s_base;
		    callgate->s_limacc = info->savedCallGate.s_limacc;
		    info->savedCallGate.s_base = 0;
		  }
	       }
	      break;
	    }
	 }
	return (0);
}

static resetTimerQueue (RciTimerQueue *tq, int size)
{
	/* Clear a timer event queue and set its size to the
	   specified value. */

	tq->size = size;
	tq->head = 0;
	tq->tail = 0;
}

static queueTimerEvent (RciTimerQueue *tq, int req, int tick)
{
	/* Queue a timer event, if there is room on the queue, by advancing 
	   the tail and adding it there. This routine can tolerate 
	   simultaneous calls to getTimerEvent() */

	int newtail;

	newtail = tq->tail + 1;
	if (newtail > tq->size) newtail = 0;
	if (newtail != tq->head)
	 { tq->buf[tq->tail].req = req;
	   tq->buf[tq->tail].tick = tick;
	   tq->tail = newtail;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static getTimerEvent (RciTimerQueue *tq, RciTimerEvent *event)
{
	/* Remove a timer event from the queue, if there are any there.
	   This routine can tolerate simultaneous calls to queueTimerEvent()*/

	if (tq->head != tq->tail)
	 { int newhead;
	   event->req = tq->buf[tq->head].req;
	   event->tick = tq->buf[tq->head].tick;
	   newhead = tq->head + 1;
	   if (newhead > tq->size) newhead = 0;
	   tq->head = newhead;
	   return (1);
	 }
	else
	 { return (0);
	 }
}
