#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#include "xydreg.h"
#include "rciioctl.h"

#define STATIC

int xyddevflag = 0;
void *xydi;

#define XYD_INTR_SPEC	0	/* device interrupt spec. */
#define XYD_REG_SPEC	0	/* register set to be mapped */
#define XYD_NULL_HANDLER (u_int(*)(caddr_t))nulldev

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		(getminor(d) & ((1<<NUM_INSTANCE_BITS)-1))
/* bit setting or-ed with the instance number to get an mmap device */
#define MMAP_MINOR		(1<<NUM_INSTANCE_BITS)
/* returns true if the minor number corresponds to an mmap device */
#define IS_MMAP_DEV(m)	(((m) & ~((1<<NUM_INSTANCE_BITS)-1)) == MMAP_MINOR)

/* Hardware-specific device routines */
static  int xydHardwareProbe (XydRegs*, dev_info_t*);
static void xydHardwareReset (XydRegs*, XydSoftc*);
static void xydHardwareIntrSetup (XydRegs*, XydSoftc*);
static void xydHardwareOpen (XydRegs*, XydSoftc*);
static void xydHardwareClose (XydRegs*, XydSoftc* );
static void xydHardwareIntrEnable (XydRegs*, XydSoftc*, int);
static  int xydHardwareIntrHandle (XydRegs*, XydSoftc*, int*);

/* Stuff relating to dynamic loading and autoconfiguration. There
   is *a lot* of boiler plate associated with this. */

static int xydidentify();
static int xydattach();
static int xyddetach();
static int xydgetinfo();
static int xydprobe();

static int xydopen(dev_t *devp, int flag, int otyp, cred_t *credp);
static int xydclose(dev_t dev, int flag, int otyp, cred_t *credp);
static int xydread(dev_t dev, struct uio *uiop, cred_t *credp);
static int xydwrite(dev_t dev, struct uio *uiop, cred_t *credp);
static int xydioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
static int xydmmap(dev_t dev, off_t off, int prot);
static int xydchpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int xydintr (caddr_t);
static u_int xydsoftintr (caddr_t);

static struct cb_ops xyd_cb_ops = {
	xydopen,
	xydclose,
	nodev,
	nodev,
	nodev,
	xydread,
	xydwrite,
	xydioctl,
	nodev,
	xydmmap,
	nodev,
	xydchpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops xyd_ops = {
	DEVO_REV,
	0,
	xydgetinfo,
	xydidentify,
	xydprobe,
	xydattach,
	xyddetach,
	nodev,
	&xyd_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"xyd driver",
	&xyd_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&xydi, sizeof(XydSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&xydi);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&xydi);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int xydidentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "xyd") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int xydgetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	XydSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(xydi, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static int xydfreeResources (dev_info_t *dip, XydSoftc *info)
{
	if ((info->resources & XYD_MMAP_MINOR_CREATED) ||
	    (info->resources & XYD_MAIN_MINOR_CREATED))
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & XYD_IHANDLER_SET)
	 { ddi_remove_intr (dip, XYD_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & XYD_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & XYD_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & XYD_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
	if (info->resources & XYD_REGS_MAPPED)
	 { ddi_unmap_regs (dip, XYD_REG_SPEC, (caddr_t*)&info->regs,
		      		   /*offset=*/0, /*len=*/0);
	 }
	ddi_soft_state_free (xydi, info->instance);
}

static int xydattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	XydSoftc *info;
	int instance;

	static char *mainNames[] = { "0", "1", "2", "3" };
	static char *mmapNames[] = { "mem0", "mem1", "mem2", "mem3" };

	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(xydi, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(xydi, instance);

	      info->dip = dip;
	      info->instance = instance;
	
	      /* map device registers */
	      if (ddi_map_regs (dip, XYD_REG_SPEC, (caddr_t*)&info->regs,
				/*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	       { xydfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYD_REGS_MAPPED;

	      /* make device quiescent */
	      xydHardwareReset (info->regs, info);

	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, XYD_INTR_SPEC,&info->iblkCookie,
		     NULL, XYD_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { xydfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "xyd intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, XYD_INTR_SPEC, info->iblkCookie);
	      info->resources |= XYD_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or pollwakeup()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, XYD_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			XYD_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { xydfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "xyd low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= XYD_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			xydsoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { xydfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= XYD_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 cmn_err (CE_CONT, "xyd: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 cmn_err (CE_CONT, "xyd: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, XYD_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, xydintr,(caddr_t)info) != DDI_SUCCESS)
	       { xydfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYD_IHANDLER_SET;

	      /* program device interrupt vector */
	      xydHardwareIntrSetup (info->regs, info);

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { xydfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYD_MAIN_MINOR_CREATED;

	      if (ddi_create_minor_node (dip, mmapNames[instance], S_IFCHR, 
		     instance|MMAP_MINOR, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { xydfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= XYD_MMAP_MINOR_CREATED;

	      ddi_report_dev(dip);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int xyddetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	XydSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (xydi, ddi_get_instance(dip));

	      /* make device quiescent */
	      xydHardwareReset (info->regs, info);
	      xydfreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static int xydprobe (dev_info_t *dip)
{
        caddr_t reg_addr;
	int status = DDI_PROBE_SUCCESS;

	/* map device registers and try to contact device */
	if (ddi_map_regs (dip, XYD_REG_SPEC, &reg_addr, 
			       /*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	 { return (DDI_PROBE_FAILURE);
	 }
	if (!xydHardwareProbe ((XydRegs*)reg_addr, dip))
	 { status = DDI_PROBE_FAILURE;
	 }
	ddi_unmap_regs (dip, XYD_REG_SPEC, &reg_addr,/*offset=*/0,/*len=*/0);
	return (status);
}

static u_int xydintr (caddr_t arg)
{
	/* Called in response to an interrupt from the XYD */

	XydSoftc *info = (XydSoftc*)arg;
	int needSoftIntr = 0;
	int pending;
	int stray;

	mutex_enter (&info->intrMux);
	pending = xydHardwareIntrHandle (info->regs, info, &stray);
	if (pending == 0 && stray == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (info->highIntrCnt < info->qsize)
	       { info->highIntrCnt++;
		 if (!info->softIntrRunning)
		  { needSoftIntr = 1;
		  }
	       }
	    }
	   else
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;		/* bump the counter */
		 mutex_exit (&info->intrMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->intrMux);
	       }
	    }
	 }
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "xyd%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
	mutex_exit (&info->intrMux);
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int xydsoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls pollwakeup and cmn-err because the high-level
	   interrupt handler can't. */

	XydSoftc *info = (XydSoftc*)arg;
	int stray;
	int cnt;

	mutex_enter (&info->lowMux);
	mutex_enter (&info->intrMux);

	/* don't know why info->softIntrRunning needs to be checked
	   in the next phrase, but it was in the manual example */
	if ((info->stray == 0 && info->highIntrCnt <= 0) ||
	    info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   mutex_exit (&info->lowMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;
	while (info->stray || info->highIntrCnt > 0)
	 { /* ASSERT (mutex_owned (&info->intrMux)); */

	   if ((stray = info->stray) != 0)
	    { info->stray = 0;
	    }
	   if ((cnt = info->highIntrCnt) > 0)
	    { info->highIntrCnt--;
	    }
	   mutex_exit (&info->intrMux);

	   if (stray)
	    { cmn_err (CE_CONT, "xyd%d: stray interrupt(s); mask=0x%x\n",
				 info->instance, stray);
	    }
	   if (cnt)
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;
		 mutex_exit (&info->lowMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->intrMux);
	       }
	    }

	   mutex_enter (&info->intrMux);
	 }
	info->softIntrRunning = 0;
	mutex_exit (&info->intrMux);
	mutex_exit (&info->lowMux);
	return (DDI_INTR_CLAIMED);	
}

STATIC int xydopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	XydSoftc *info;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
	if (!info->regs)	/* paranoid; this check shouldn't be needed */
	 { return (ENXIO);
	 }

	/* now determine whether this is a main device or an mmap device.
	   Enforce exclusive open for main devices */
	mutex_enter (info->muxp);
	if (IS_MMAP_DEV(getminor(*dev)))
	 { /* OK, just return */
	 }
	else
	 { if (info->flags & XYD_BUSY)
	    { mutex_exit (info->muxp);
	      return (EBUSY);
	    }
	   else
	    { info->flags |= XYD_BUSY;
	    }
	   info->qsize = TQ_DEFAULT_SIZE;
	   info->intrCnt = 0;
	   bzero ((char*)&info->pollhead, sizeof(struct pollhead));

	   xydHardwareOpen (info->regs, info);
	 }
	mutex_exit (info->muxp);
	return (0);
}

static void doclose (XydSoftc *info)
{
	xydHardwareClose (info->regs, info);
	info->flags = 0;
}

STATIC int xydclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	XydSoftc *info;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (0);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & XYD_BUSY) )
	 { mutex_exit (info->muxp);
	   return (0);
	 }
	mutex_exit (info->muxp);
	doclose (info);
	return (0);
}

STATIC int xydread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns a single byte (set to 1) if an interrupt
	   arrived. */

	XydSoftc *info;
	int cnt;
	int ack = 1;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & XYD_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	if (cnt)
	 { info->intrCnt--;
	 }
	mutex_exit (info->muxp);

	if (cnt)
	 { if (ureadc (ack, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int xydwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int xydmmap (dev_t dev, off_t off, int prot)
{
	XydSoftc *info;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(dev))) == NULL)
	 { return (-1);
	 }
	if (off < 0 || off >= sizeof(XydRegs))
	 { return (-1);
	 }
	return (hat_getkpfnum((char*)info->regs + off));
}

STATIC int xydchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	/* Check to see if any interrupt events have arrived. If not,
	   then queue the process for a possible select wakeup. */

	XydSoftc *info;
	int cnt;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & XYD_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	mutex_exit (info->muxp);

	/* not really sure how system prevents races with the interrupt
	   handler here, but this is what the example says to do ... */
	if ((events & (POLLIN|POLLRDNORM)) && cnt > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = &info->pollhead;
	    }
	 }
	return (0);
}

STATIC int xydioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	XydSoftc *info;
	int isMmapDev;

	if ((info = ddi_get_soft_state(xydi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if ((isMmapDev = IS_MMAP_DEV(getminor(dev))) == 0)
	 { /* check for soft-close */
	   mutex_enter (info->muxp);
	   if (! (info->flags & XYD_BUSY) )
	    { mutex_exit (info->muxp);
	      return (EBADF);
	    }
	   mutex_exit (info->muxp);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_INTR_ENABLE:
	    { mutex_enter (info->muxp);
	      xydHardwareIntrEnable (info->regs, info, (int)arg);
	      mutex_exit (info->muxp);
	      break;
	    }
	   case RCIIOC_SET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = (int)arg;
	      if (size <= 0)
	       { return (EINVAL);
	       }
	      else
	       { mutex_enter (info->muxp);
		 info->qsize = size;
		 info->intrCnt = 0;
		 mutex_exit (info->muxp);
	       }
	      break;
	    }
	   case RCIIOC_GET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = info->qsize;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
/* RCIIOC_SOFT_CLOSE is a special hack. On some machines, calling
   close(2) does not always result in the driver close routine being 
   called right away. So we fake it with an ioctl, just to be sure.
   It is the responsiblity of the application to call close(2) directly
   after the SOFT_CLOSE. Once SOFT_CLOSE has been called, the driver 
   will allow another exclusive open on the main device, and so concurrent
   accesses by the original descriptor are possible (but probably not
   desirable). */   
	   case RCIIOC_SOFT_CLOSE:
	    { if (isMmapDev)
	       { return (ENOTTY);
	       }
	      doclose (info);      
	      break;
	    }
	 }
	return (0);
}

static unsigned char idcheck[] = "VMEIDXYC240    1 11 ";
static int xydHardwareProbe (XydRegs *regs, dev_info_t *dip)
{
	/* Verify that the board is really an XVME-240, using the
	   ID signature built in to the board. */

	char c;
	int i;

	for (i=0;i<20;i++)
	 { if ((ddi_peekc (dip, (char*)&regs->id[i+i+1], &c) == DDI_FAILURE) )
	    { cmn_err (CE_CONT, "xyd: no memory at device address\n");
	      return(0);
	    }
	   if (c != idcheck[i])
	    { cmn_err (CE_CONT, "xyd: bad signature at device address\n");
	      return(0);
	    }
	 }
	return(1);
}

static void xydHardwareReset (XydRegs *regs, XydSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	regs->intmask = 0; /* disables interrupts */
	regs->csr = XYD_RESET;
	regs->csr = 0;
	regs->csr = XYD_RED_OFF | XYD_GREEN_OFF;
	regs->dirs = XYD_DIRMASK;
	regs->intclear = XYD_INTRCLEAR;
}

static void xydHardwareIntrSetup (XydRegs* regs, XydSoftc* info)
{
	/* if the device interrupt vector or IPL is programmable,
	   do this here based on info provided in info->idevCookie. */
	info->regs->vector = info->idevCookie.idu._idev_vector;
}

static void xydHardwareClose(XydRegs *regs, XydSoftc *info)
{
	/* make the device quiet when it is closed */

	regs->intmask = 0; /* disables interrupts, so more need for mutex */
	regs->intclear = XYD_INTRCLEAR;
	regs->csr = XYD_RED_OFF | XYD_GREEN_OFF;
}

static void xydHardwareOpen(XydRegs *regs, XydSoftc *info)
{
	/* initialize the device when it is openned */	

	regs->intclear = XYD_INTRCLEAR;
	regs->dirs = XYD_DIRMASK;
}

static void xydHardwareIntrEnable(XydRegs *regs, XydSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	if (enable)
	 { regs->intmask |= XYD_INTRMASK;
	   regs->intclear = XYD_INTRCLEAR;
	   if (! (regs->csr & XYD_INTENB))
	    { regs->csr |= XYD_INTENB;
	    }
	 }
	else
	 { regs->csr &= ~XYD_INTENB;
	   regs->intmask = 0;
	   regs->intclear = XYD_INTRCLEAR;
	 }
}

static int xydHardwareIntrHandle(XydRegs *regs, XydSoftc *info, int *stray)
{
	/* clear interrupts, return true if a valid interrupt is pending,
	   and return a mask of any stray interrupts in *stray. */
	int pending;

	pending = regs->intpend;	
	regs->intclear = pending;
	if (info->flags & XYD_BUSY)
	 { *stray = (pending & ~XYD_INTRMASK);
	   return (pending & XYD_INTRMASK);
	 }
	else
	 { /* device isn't open; therefore, any interrupt is a stray */
	   *stray = pending;
	   return (0);
	 }
}
